"""Kat script exceptions.

These are special exceptions that add extra context-specific information when errors
occur during parsing or unparsing.
"""

from itertools import chain
from collections import defaultdict
from ..exceptions import FinesseException
from ..utilities import ngettext, option_list, add_linenos
from .containers import KatCoordinate, KatBounds
from .util import index_ranges, scriptsorted


class KatParsingError(FinesseException):
    """An error during the parsing process.

    This handles errors raised by the tokenizer, parser and builder. These are,
    confusingly, collectively referred to as the "parser".
    """


class KatScriptError(KatParsingError):
    """Detailed kat script error.

    Parameters
    ----------
    error : str or :py:class:`Exception`
        The error or error message.

    container : :class:`.KatAddressable`
        The kat script container, used to show lines before/after error lines if
        requested.

    error_items : sequence of sequences of :class:`.ScriptLocation`
        The error(s) to show. Each item of the topmost sequence should itself be an
        ordered sequence of :class:`.ScriptLocation` objects in increasing order of
        depth. The last item is considered to be where the error occurred, but the start
        and end lines (plus optional buffer) of each of the other items are also shown.
    """

    # Maximum number of unrelated lines to show before and after show items.
    MAX_BUFFER_LINES = 1

    def __init__(self, error, container, error_items):
        # Store passed data as attributes in case another exception wants to grab it.
        self.error_msg = error
        self.container = container
        self.error_items = error_items
        super().__init__(self.message())

    def _dedup_bounds(self, bounds):
        """Sequence of bounds without overlapping intervals.

        Assumes `bounds` are sorted.
        """
        dedup_bounds = [bounds[0]]

        for current in bounds[1:]:
            previous = dedup_bounds[-1]
            if current.start <= previous.stop:
                dedup_bounds[-1] = previous.rexpand(current.stop)
            else:
                dedup_bounds.append(current)

        return dedup_bounds

    @property
    def _marker_items(self):
        return [items[-1] for items in self.error_items]

    def _marker_bounds(self):
        """Sequence of sorted marker coordinates without overlapping intervals.

        The marker bounds are those of the final item in each item of
        :attr:`.error_items`.
        """
        return self._dedup_bounds(
            scriptsorted(set(item.bounds for item in self._marker_items))
        )

    @property
    def _flattened_error_items(self):
        return chain.from_iterable(self.error_items)

    def _flattened_error_bounds(self):
        return self._dedup_bounds(
            scriptsorted(set([item.bounds for item in self._flattened_error_items]))
        )

    @property
    def error_lines(self):
        lines = set()
        for bounds in self._flattened_error_bounds():
            lines.add(bounds.start.lineno)
            lines.add(bounds.stop.lineno)
        return list(lines)

    def _show_error_bounds(self):
        """Error items with extra lines before and after."""
        show_bounds = []

        for bounds in self._flattened_error_bounds():
            # Some buffer before and after the error, if available in the container.
            start = max(
                KatCoordinate(bounds.start.lineno - self.MAX_BUFFER_LINES, 1),
                self.container.start,
            )
            stop = min(
                KatCoordinate(bounds.stop.lineno + self.MAX_BUFFER_LINES + 1, 0),
                self.container.stop,
            )

            show_bounds.append(bounds.lexpand(start).rexpand(stop))

        return self._dedup_bounds(show_bounds)

    def items_to_linemarkers(self, items):
        linemarkers = defaultdict(lambda: defaultdict(lambda: False))

        for item in items:
            if item.start.lineno != item.stop.lineno:
                raise NotImplementedError(
                    "cannot draw markers on error items spanning multiple lines"
                )

            for index in range(item.start.index, item.stop.index):
                linemarkers[item.start.lineno][index] = True

        return linemarkers

    def rubric(self):
        if not (error_msg := self.error_msg):
            error_msg = ""
        error_msg = str(error_msg)

        lines = self.error_lines

        if len(lines) == 1:
            rubric = f"line {lines[0]}: {error_msg}"
        else:
            rubric = f"lines {option_list(index_ranges(lines), 'and')}: {error_msg}"

        return rubric.strip()

    def message(self):
        msglines = defaultdict(str)
        markers = defaultdict(lambda: defaultdict(lambda: False))

        for lineno, linechunks in self.chunkify().items():
            for bounds, is_error in linechunks:
                if bounds.isempty():
                    # Skip empty bounds, which might otherwise cause an empty extra line
                    # to appear at the end of the error.
                    continue

                # Grab the first line (excluding newline character) since the bounds
                # don't cross lines.
                script = self.container.script_lines(bounds)
                assert len(script) < 2
                if script:
                    msglines[lineno] += script[0]

                if is_error:
                    for index in range(bounds.start.index, bounds.stop.index):
                        markers[lineno][index] = True

        msgwithlines = add_linenos(msglines.keys(), msglines.values())
        lines = []
        if msgwithlines:
            # Work out how wide the line column is. There is +3 for the prefix.
            lcolbuf = len(msgwithlines[0]) - len(next(iter(msglines.values()))) + 3

            # Add arrows on the error lines and markers under the errors.
            lastlineno = None
            for lineno, line in zip(msglines, msgwithlines):
                if lastlineno:
                    diff = lineno - lastlineno
                    if diff > 1:
                        missingmsg = ngettext(
                            diff - 1, "%d skipped line", "%d skipped lines"
                        )
                        lines.append(f"{' ' * lcolbuf}*** {missingmsg} ***")

                prefix = "-->" if markers[lineno] else "   "
                lines.append(prefix + line)
                linemarks = ""
                if markers[lineno]:
                    for index in range(1, len(line) + 1):
                        if markers[lineno][index]:
                            linemarks += "^"
                        else:
                            linemarks += " "
                    linemarks = linemarks.rstrip()
                    if linemarks:
                        lines.append(f"{' ' * lcolbuf}{linemarks}")

                lastlineno = lineno

        lines.insert(0, self.rubric())

        return "\n".join(lines)

    def chunkify(self):
        """Convert the error script into a dict of line numbers to (str, is_error)
        sequences.

        Each (str, is_error) chunk contains a piece of the script specified in
        `error_items`, with the `is_error` flag set depending on whether each error item
        appeared last in a `error_items` sequence, which specifies that it is an error
        as opposed to just useful related code.

        Returns
        -------
        :class:`dict`
            Mapping of line numbers to lists of (:class:`str`, :class:`bool`) tuples
            indicating script chunks and whether they contain an error.
        """
        # Merge the error and marker bounds by starting coordinate, adding an extra flag
        # to remember the type.
        bounds = sorted(
            [(bound, False) for bound in self._show_error_bounds()]
            + [(bound, True) for bound in self._marker_bounds()],
            key=lambda chunk: chunk[0].start,
        )

        merged = [bounds[0]]

        for current_bound, current_flag in bounds[1:]:
            previous_bound, previous_flag = merged[-1]

            # Merge bounds if necessary, with marker intervals taking priority.
            if current_bound == previous_bound:
                # Bounds entirely overlap.
                merged[-1] = (previous_bound, previous_flag or current_flag)
            elif current_bound.start < previous_bound.stop:
                # Bounds overlap.
                inner1 = current_bound.start
                merged[-1] = (previous_bound.rcontract(inner1), previous_flag)

                if current_bound.stop <= previous_bound.stop:
                    inner2 = current_bound.stop
                    merged.append((KatBounds(inner1, inner2), current_flag))
                    merged.append((previous_bound.lcontract(inner2), previous_flag))
                else:
                    merged.append((previous_bound.lcontract(inner1), current_flag))
            else:
                merged.append((current_bound, current_flag))

        linebounds = defaultdict(list)

        # Split any chunks spanning multiple lines.
        for bound, flag in merged:
            if bound.start.lineno == bound.stop.lineno:
                linebounds[bound.start.lineno].append((bound, flag))
            else:
                laststart = bound.start
                for lineno in range(bound.start.lineno, bound.stop.lineno):
                    newstop = KatCoordinate(laststart.lineno + 1, 0)
                    linebounds[laststart.lineno].append(
                        (KatBounds(laststart, newstop), flag)
                    )
                    laststart = newstop
                # Last line.
                linebounds[bound.stop.lineno].append(
                    (KatBounds(KatCoordinate(bound.stop.lineno, 0), bound.stop), flag)
                )

        return linebounds


class KatSyntaxError(KatScriptError):
    """Error with kat syntax."""

    def __init__(self, message, script, token):
        super().__init__(message, script, [[token]])


class KatMissingAfterDirective(KatSyntaxError):
    """Error for when an element name or function '(' is missing."""

    def __init__(self, container, directive):
        self.directive = directive
        super().__init__("missing '(' or element name", container, directive)



import pytest
import finesse
import numpy as np
from numpy.testing import assert_allclose

@pytest.fixture
def model_m():
    kat = finesse.Model()
    kat.parse(
    """
    l l1
    m m1 R=0.5 T=0.5
    s s1 m1.p2 m2.p1
    m m2 R=0 T=1
    l l2

    link(l1, m1)
    link(l2, m2)

    pd P1 l1.p1.i
    pd P2 l2.p1.i
    """
    )
    return kat


@pytest.fixture
def model_bs():
    kat = finesse.Model()
    kat.parse_legacy(
        """
    l i1 1 0 0 nlaser
    s sin 0 nlaser nBS1

    bs BS 0.5 0.5 0 0 nBS1 nBS2 nBS3 nAS
    s sX 0 1 nBS3 nX
    s sY 0 1 nBS2 nY

    m mX 1 0 0 nX dump
    m mY 1 0 0 nY dump

    pd PIN nlaser*
    pd PREFL nlaser
    pd PAS nAS

    xaxis* BS phi lin 0 90 10
    yaxis lin abs:deg

    maxtem off
    """
    )
    return kat

@pytest.mark.parametrize("f",(0, 0.001, 1e15))
@pytest.mark.parametrize("angle",(0, -89, 45))
@pytest.mark.parametrize("nr",(1, 1.45))
@pytest.mark.parametrize("phi",(0, 1, 1130123.3))
@pytest.mark.parametrize("flag",(True, False))
def test_bs_power_conservation(model_bs, f, angle, nr, phi, flag):
    model_bs.phase_config.V2_TRANSMISSION_PHASE = flag
    model_bs.i1.f = f
    model_bs.BS.alpha = angle
    model_bs.BS.phi = phi
    model_bs.spaces.sX.nr = nr
    model_bs.spaces.sY.nr = nr
    out = model_bs.run()
    assert_allclose(out["PREFL"] + out["PAS"], out["PIN"], rtol=1e-14, atol=1e-14)


@pytest.mark.parametrize("f",(0, 0.001, 1e15))
@pytest.mark.parametrize("nr",(1, 1.45))
@pytest.mark.parametrize("phi1",(0, 1, 2*np.pi, -2*np.pi))
@pytest.mark.parametrize("phi2",(0, 1, 2*np.pi, -2*np.pi))
@pytest.mark.parametrize("flag",(True, False))
def test_mirror_power_conservation(model_m, f, nr, phi1, phi2, flag):
    model_m.phase_config.V2_TRANSMISSION_PHASE = flag
    model_m.l1.f = f
    model_m.l2.f = f
    model_m.m1.phi = phi1
    model_m.m2.phi = phi2
    model_m.spaces.s1.nr = nr
    out = model_m.run()
    assert_allclose(out['P1'] + out['P2'], 2, rtol=1e-14, atol=1e-14)

@pytest.mark.xfail(reason="numerical precision problems with large phi values")
@pytest.mark.parametrize("f",(0, 0.001, 1e15))
@pytest.mark.parametrize("nr",(1, 1.45))
@pytest.mark.parametrize("phi1",(1130123000.3,))
@pytest.mark.parametrize("phi2",(1130123000.3,))
@pytest.mark.parametrize("flag",(False, ))
def test_mirror_power_conservation_large_phi(model_m, f, nr, phi1, phi2, flag):
    model_m.phase_config.V2_TRANSMISSION_PHASE = flag
    model_m.l1.f = f
    model_m.l2.f = f
    model_m.m1.phi = phi1
    model_m.m2.phi = phi2
    model_m.spaces.s1.nr = nr
    out = model_m.run()
    assert_allclose(out['P1'] + out['P2'], 2, rtol=1e-14, atol=1e-14)


"""Tokenization for kat script.

This differs from a typical tokenizer in that it tokenizes whitespace and
newline characters, which have important meanings in kat script and help
maintain formatting in later model serialization.

Inspired by the Python standard library `tokenize` module and lex.py from
David Beazley's SLY (https://github.com/dabeaz/sly/).

Sean Leavey <sean.leavey@ligo.org>
"""
import re
from io import StringIO
from .tokens import (
    WHITESPACE,
    NEWLINE,
    COMMENT,
    NONE,
    BOOLEAN,
    REFERENCE,
    NAME,
    NUMBER,
    STRING,
    LITERALS,
)
from .containers import (
    KatToken,
    KatMetaToken,
    KatWhitespaceToken,
    KatNoneToken,
    KatBooleanToken,
    KatReferenceToken,
    KatNumberToken,
    KatStringToken,
    KatFile,
)
from .exceptions import KatSyntaxError


class KatTokenizer:
    """Kat script token generator.

    Regular expressions define patterns to match prototypes. Prototypes are then either
    emitted as tokens as-is, or modified by callback methods defined in this class.

    Parameters
    ----------
    graph : :class:`.KatGraph`, optional
        The parse graph to add parsed tokens to. Defaults to a new :class:`.KatGraph`.
    """

    # All token rules, in order of matching precedence.
    _TOKEN_RULES = {
        "WHITESPACE": WHITESPACE,
        "NEWLINE": NEWLINE,
        "COMMENT": COMMENT,
        "NONE": NONE,  # Has to be above NAME.
        "BOOLEAN": BOOLEAN,
        "REFERENCE": REFERENCE,  # Has to be above NAME.
        "NUMBER": NUMBER,  # Has to be above NAME (since it can match "inf")
        "STRING": STRING,
        "NAME": NAME,
        **{key: re.escape(value) for key, value in LITERALS.items()},
    }

    # Token constructors.
    _TOKEN_CLASSES = {
        "WHITESPACE": KatWhitespaceToken,
        "NONE": KatNoneToken,
        "BOOLEAN": KatBooleanToken,
        "REFERENCE": KatReferenceToken,
        "NUMBER": KatNumberToken,
        "STRING": KatStringToken,
    }

    def __init__(self):
        self.done = None
        self.script = None
        self._lineno = None
        self._index = None
        self._matcher = None
        self._nesting = None
        self._build_rules()

    def _build_rules(self):
        """Create overall expression with named items."""
        rules = "|".join(
            [f"(?P<{name}>{pattern})" for name, pattern in self._TOKEN_RULES.items()]
        )
        self._matcher = re.compile(rules)

    def _raise_error(self, error_msg, token):
        raise KatSyntaxError(error_msg, self.script, token)

    def _raise_lexing_error(self, token):
        self._raise_error(f"illegal character {token.value[0]!r}", token)
        # Skip over the bad token.
        self._index += len(token.value)
        return token

    def tokenize(self, string):
        """Tokenize specified `string`.

        Parameters
        ----------
        string : :class:`str`
            The string to tokenize kat script from.

        Yields
        ------
        :class:`.KatToken`
            The next token read from `string`.
        """
        yield from self.tokenize_file(StringIO(string))

    def tokenize_file(self, fobj):
        """Generate tokens from the given file.

        Parameters
        ----------
        fobj : :class:`io.FileIO`
            The file object to tokenize kat script from. This should be opened in text
            mode.

        Yields
        ------
        :class:`.KatToken`
            The next token read from `fobj`.
        """
        # Reset tokenizer state.
        self.done = False
        self.errors = []
        self.script = KatFile("")
        self._nesting = []
        self._lineno = 1
        self._index = 0
        previous_line = ""
        line = ""

        while True:
            previous_line = line
            line = fobj.readline()

            # Lines are only empty at the end of the file (blank lines are "\n").
            if not line:
                break

            linelength = len(line)
            # Store the line for use in error messages.
            self.script.add(line)

            # The current string index on the current line. The start/stop values stored
            # in tokens are this value + 1.
            # This can be modified by the error function, so it's part of the object
            # scope. We also remember the total extra space characters used to
            # compensate for tabs, which only take up 1 character in the file but more
            # when displayed.
            self._index = 0
            self._taboffset = 0

            while self._index < linelength:
                matches = self._matcher.match(line, self._index)

                if matches:
                    raw_value = matches.group()

                    # Compensate for tabs. Tabs only take up one character but when
                    # displayed take up whatever we decide here.
                    start_index = self._index + self._taboffset + 1
                    self._taboffset += (
                        KatWhitespaceToken.TAB_SIZE - 1
                    ) * raw_value.count("\t")
                    stop_index = matches.end() + self._taboffset + 1

                    token_type = matches.lastgroup
                    # If a special token type is defined, use its corresponding class
                    # instead of :class:`.KatToken`.
                    token_class = self._TOKEN_CLASSES.get(token_type, KatToken)

                    try:
                        token = token_class(
                            lineno=self._lineno,
                            start_index=start_index,
                            stop_index=stop_index,
                            type=token_type,
                            raw_value=raw_value,
                        )
                    except SyntaxError as e:
                        # There was an error determining the final value of the token.
                        self._raise_error(
                            e,
                            KatToken(
                                self._lineno,
                                start_index,
                                stop_index,
                                token_type,
                                raw_value,
                            ),
                        )

                    if token_callback := getattr(self, f"on_{token_type}", False):
                        token_callback(token)

                    self._index = matches.end()  # Updates index for next token.
                else:
                    # A lexing error.
                    token = KatToken(
                        self._lineno,
                        start_index=self._index + 1,
                        stop_index=linelength + 1,
                        type="ERROR",
                        raw_value=line[self._index :],
                    )
                    # Leave it to the error function to advance the index and/or recover
                    # the token.
                    token = self._raise_lexing_error(token)

                yield token

        if self._nesting:
            # Unclosed parenthesis/parentheses.
            for token in self._nesting:
                self._raise_error(f"unclosed '{token.value}'", token)

        # Add an implicit NEWLINE if the input doesn't end in one (this simplifies the
        # parser rules).
        if not previous_line or (previous_line and previous_line[-1] not in "\r\n"):
            yield KatMetaToken(
                self._lineno, self._index + 1, self._index + 1, "NEWLINE"
            )
            self._lineno += 1
            self._index = 0

        yield KatMetaToken(self._lineno, 1, 1, "ENDMARKER")
        self.done = True

    def on_NEWLINE(self, token):
        self._lineno += len(token.value)

    def on_LBRACKET(self, token):
        self._nesting.append(token)

    def on_RBRACKET(self, token):
        try:
            assert self._nesting.pop().value == "["
        except (IndexError, AssertionError):
            self._raise_error("extraneous ']'", token)

    def on_LPAREN(self, token):
        self._nesting.append(token)

    def on_RPAREN(self, token):
        try:
            assert self._nesting.pop().value == "("
        except (IndexError, AssertionError):
            self._raise_error("extraneous ')'", token)



import finesse
from finesse.cmatrix import KLUMatrix
import numpy as np
import pytest

def test_fill_za_zmvc():
    M = KLUMatrix('M')
    M.add_diagonal_elements(3, 0, 'a')
    M.add_diagonal_elements(3, 1, 'b')
    M.add_diagonal_elements(1, 2, 'c')

    Vac = M.get_sub_matrix_view(0, 2, 'a->c', False)
    Vbc = M.get_sub_matrix_view(1, 2, 'b->c', True)

    Vca = M.get_sub_matrix_view(2, 0, 'c->a', False)
    Vcb = M.get_sub_matrix_view(2, 1, 'c->b', True)

    M.construct()

    I = np.array([
        [1,2j,3],
        [4,5,6j],
        [7,8j,9]], dtype=np.complex128)
    V = np.array([1,2j,3], dtype=np.complex128)

    Vac.do_fill_za_zmvc(3j, I, V)
    assert((abs(Vac.view - 3j*(I@V)) < 1e-15).all())

    Vbc.do_fill_za_zmvc(3j, I, V)
    assert((abs(Vbc.view - (3j*I@V).conj()) < 1e-15).all())

    Vca.do_fill_za_zmvc(3j, I, V)
    assert((abs(Vca.view.T - 3j*I@V) < 1e-15).all())

    Vcb.do_fill_za_zmvc(3j, I, V)
    assert((abs(Vcb.view.T - (3j*I@V).conj()) < 1e-15).all())



import os

__all__ = ["list_styles", "use", "context"]

# Get style directory and available styles
style_dir = os.path.split(os.path.realpath(__file__))[0]


def list_styles():
    """Lists all the styles (``.mplstyle`` files) available
    to use.

    Returns
    -------
    list
        A list containing the names of all available styles.
    """
    return [
        os.path.splitext(f)[0] for f in os.listdir(style_dir) if f.endswith(".mplstyle")
    ]


def get_style_path(style):
    """Returns the path to each style passed as an argument.

    Parameters
    ----------
    style : str or list of str
        The style(s) to get the paths of

    Returns
    -------
    list of str
        A list containing the paths of all requested styles.
    """
    if isinstance(style, str) or hasattr(style, "keys"):
        # If name is a single str or dict, make it a single element list.
        styles = [style]
    else:
        styles = style
    styles = list(map((lambda s: os.path.join(style_dir, s + ".mplstyle")), styles))
    return styles


def use(style):
    """Sets :mod:`matplotlib.pyplot` parameters to use the given style type.

    See :func:`.list_styles` for a list of all available styles.

    Parameters
    ----------
    style : str
        Name of the style to use.

    Examples
    --------
    To set-up your plots to use the ``default.mplstyle`` style-sheet, simply
    write::

        import finesse.plotting as fplt
        fplt.use('default')
    """
    import matplotlib.pyplot as plt

    plt.style.use(get_style_path(style))


def context(style, after_reset=False):
    """Sets :mod:`matplotlib.pyplot` parameters to use the given style type
    within the current context.

    See :func:`.list_styles` for a list of all available styles.

    Parameters
    ----------
    style : str
        Name of the style to use.

    after_reset : bool
        If True, apply style after resetting settings to their defaults;
        otherwise, apply style on top of the current settings.

    Examples
    --------
    To use the ``default.mplstyle`` style-sheet temporarily, use the `with`
    statement as so::

        import finesse.plotting as fplt
        with fplt.context('default'):
            ...
    """
    import matplotlib.pyplot as plt

    return plt.style.context(get_style_path(style), after_reset)



"""Configuration tools."""

# NOTE: this module gets imported by `finesse` directly, so cannot itself import from
# `finesse` and cannot import packages that themselves import from `finesse`.
import os
from pathlib import Path
import importlib.resources
from configparser import RawConfigParser
import logging
from fnmatch import fnmatch

from . import datastore
from .utilities import option_list


_PACKAGE_LOGGER = logging.getLogger(__package__)


def config_instance():
    """The Finesse configuration object for the current session.

    Returns
    -------
    :class:`configparser.RawConfigParser`
        The Finesse configuration object.
    """
    return datastore.init_singleton(_FinesseConfig)


def configure(
    plotting=False, log_level=None, log_exclude=None, jupyter_tracebacks=None
):
    """Configure Finesse runtime options.

    Parameters
    ----------
    plotting : bool, optional
        Initialise Finesse plot theme for display.

    log_level : str, optional
        Configure a special :mod:`finesse` log handler to print log messages emitted at
        this level or higher to the error stream. The Python log levels "debug", "info",
        "warning", "error" and "critical" are supported, as are their corresponding
        level numbers (see :mod:`logging`). The logger is only created once, but
        subsequent calls to this function will update the log level.

    log_exclude : list of :class:`str`, optional
        Names of log channels to suppress, if a Finesse log handler can be configured
        (see notes).

    jupyter_tracebacks : bool, optional
        Show full tracebacks in Finesse errors when using Finesse in IPython. This
        setting does not work reliably in other environments.

    Notes
    -----
    The `log_level` and `log_exclude` parameters are only set if the root `finesse` log
    handler is a `_FinesseStreamHandler`. This is the case when the user has not
    configured any other `finesse` log handler prior to running this function for the
    first time in the current session, which allows this function to safely create one.
    If another log handler does exist and `log_level` and/or `log_exclude` are set, they
    are ignored and a warning is emitted.

    See Also
    --------
    :func:`finesse.plotting.tools.init`
    """
    # NOTE: Before modifying this function, note that it may be called multiple times
    # during the execution of Finesse and should therefore remain idempotent.

    if plotting:
        from .plotting import init as init_plotting

        init_plotting(mode="display")

    # Remove any existing handlers configured by this function.
    _clear_log_handler_instance()

    if log_level is not None or log_exclude is not None:
        try:
            handler = log_handler_instance()
        except ExistingLogHandlerError:
            import warnings

            warnings.warn(
                f"an existing log handler was configured for '{__package__}' prior to "
                f"the first call to this function; refusing to set log level or "
                f"excludes"
            )
        else:
            try:
                log_level = log_level.upper()
            except AttributeError:
                # Probably a number.
                pass

            if log_level:
                # Set the root logger's level, not the handler.
                _PACKAGE_LOGGER.setLevel(log_level)

            if log_exclude:
                for exclude in log_exclude:
                    handler.exclude(exclude)

    if jupyter_tracebacks is not None:
        from .environment import show_tracebacks

        show_tracebacks(jupyter_tracebacks)


def autoconfigure():
    """Automatically configure runtime options based on the environment."""
    from .environment import is_interactive

    if is_interactive():
        # The user has imported Finesse inside a notebook or similar interactive
        # session. Configure some aspects of Finesse automatically: pretty plots,
        # suppress tracebacks and set logging to emit warnings and above to stderr if
        # the user hasn't already configured a logger prior to importing.
        log_level = "warning" if not _PACKAGE_LOGGER.handlers else None
        configure(plotting=False, log_level=log_level, jupyter_tracebacks=False)
    else:
        # Use the defaults.
        configure()


def log_handler_instance():
    """Get the Finesse stream handler instance, creating and adding it to the root
    logger if necessary.

    Raises
    ------
    :class:`.ExistingLogHandlerError`
        If an existing log handler is set for the `finesse` logger by the time this
        function is first called in the current session.
    """

    def instance():
        """The root Finesse log handler for the current session."""
        return datastore.init_singleton(_FinesseStreamHandler)

    # Get the Finesse stream handler, if necessary.
    if not datastore.has_singleton(_FinesseStreamHandler):
        if _PACKAGE_LOGGER.handlers:
            raise ExistingLogHandlerError()
        else:
            # Configure the Finesse stream handler.
            handler = instance()
            handler.setFormatter(
                logging.Formatter(config_instance()["logging"]["log_format"])
            )
            _PACKAGE_LOGGER.addHandler(handler)

    return instance()


def _clear_log_handler_instance():
    """Remove existing :class:`._FinesseStreamHandler` from the root Finesse logger if
    present."""

    for handler in list(_PACKAGE_LOGGER.handlers):
        if type(handler) is _FinesseStreamHandler:
            # Remove.
            _PACKAGE_LOGGER.removeHandler(handler)

    try:
        datastore.invalidate_singleton(_FinesseStreamHandler)
    except KeyError:
        # No existing instance.
        pass


class _FinesseStreamHandler(logging.StreamHandler):
    """Finesse stream handler.

    This class provides a mechanism to exclude displayed log channels by wildcard. It is
    otherwise identical to :py:class:`logging.StreamHandler`.

    Do not use this class directly. It is an internal class handled by :func:`configure`
    and instances may be managed and deleted at its whim.
    """

    def __init__(self):
        super().__init__()
        self.__excluded_channels = None
        self.reset_exclude_patterns()

    def exclude(self, pattern):
        self.__excluded_channels.add(pattern)

    def reset_exclude_patterns(self):
        """Empty the configured log channel exclude patterns, and return what was
        there."""
        old_excludes = set(self.__excluded_channels or [])
        self.__excluded_channels = set()
        return old_excludes

    def filter(self, record):
        for pattern in self.__excluded_channels:
            if fnmatch(record.name, pattern):
                # Skip the record.
                return
        return record


class _FinesseConfig(RawConfigParser):
    """The built-in and user configuration for Finesse.

    Do not instantiate this class directly; use :func:`config_instance`.
    """

    # Order in which user configs are loaded.
    _USER_CONFIG_LOAD_ORDER = ["user_config_path", "cwd_config_path"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.write_user_config()
        self._load_finesse_configs()

    @classmethod
    def user_config_paths(cls):
        return {
            config: getattr(cls, config)() for config in cls._USER_CONFIG_LOAD_ORDER
        }

    @classmethod
    def user_config_path(cls):
        return cls.user_config_dir() / "usr.ini"

    @classmethod
    def cwd_config_path(cls):
        return Path.cwd() / "finesse.ini"

    @classmethod
    def user_config_dir(cls):
        r"""The path to the user's config directory for Finesse.

        The exact path is determined by the current platform and the presence of certain
        environment variables:

        .. rubric:: Windows

        A folder called ``finesse`` in the folder pointed to by the environment variable
        ``%APPDATA`` (usually ``%HOMEPATH%\AppData\Roaming``).

        .. rubric:: POSIX (including Mac OS X and WSL)

        A directory called ``finesse`` inside either the path pointed to by the
        environment variable ``XDG_CONFIG_HOME`` or, if that value cannot be found or is
        empty, ``~/.config``.

        Returns
        -------
        :py:class:`pathlib.Path` or None
            The path to the Finesse config directory.

        Raises
        ------
        :py:class:`RuntimeError`
            If no config directory can be determined.
        """
        from .environment import IS_WINDOWS

        if IS_WINDOWS:
            try:
                config_dir = Path(os.environ["APPDATA"])
            except KeyError:
                # NOTE: we assume %APPDATA% always exists on any normal Windows machine,
                # which should be the case. If it's not, we might need to change Finesse
                # to handle having no user config path.
                raise RuntimeError(
                    r"The %APPDATA% environment variable is required for Finesse to "
                    r"store user configuration, but it was not found. Please ensure "
                    r"this environment variable exists in the environment in which "
                    r"Finesse is being run."
                )
        else:
            # Path.home() raises RuntimeError if no home is found.
            config_dir = Path(
                os.environ.get("XDG_CONFIG_HOME", Path.home() / ".config")
            )

        return config_dir / "finesse"

    @classmethod
    def write_user_config(cls, force=False):
        """Copy the default config files to the user's config directory."""
        logger = logging.getLogger(__name__)
        user_config_path = cls.user_config_path()

        if force or not user_config_path.is_file():
            # Copy barebone user config file contents into user's user config path.
            logger.info(f"Writing user config file to {user_config_path}.")

            user_config_path.parent.mkdir(parents=True, exist_ok=True)
            with user_config_path.open("wb") as fobj:
                fobj.write(importlib.resources.read_binary(__package__, "usr.ini.dist"))

    def _load_finesse_configs(self):
        """Read the built-in and any user configuration files.

        The built-in configuration is loaded first, then the user configuration files
        are loaded in the order specified in :attr:`.USER_PATHS`. This means user
        configuration options can overwrite built-in options, and options from later
        paths in :attr:`.USER_PATHS` can overwrite options from earlier paths.
        """
        user_config_paths = self.user_config_paths().values()

        # Load the bundled configuration.
        self.read_string(
            importlib.resources.read_text(__package__, "finesse.ini"),
            source="<bundled finesse.ini>",
        )

        # Parse all configurations, from lowest to highest priority.
        parsed = self.read(user_config_paths)

        if not parsed:
            paths = option_list(user_config_paths)

            raise ConfigNotFoundError(f"Could not find user config files at {paths}.")


class ConfigNotFoundError(Exception):
    """Indicates a Finesse configuration could not be loaded."""


class ExistingLogHandlerError(Exception):
    """Indicates an existing non-Finesse log handler was found registered for the
    Finesse package."""



"""
Tests that a mirror signal scatters HOM correctly
"""
import pytest
import finesse
import numpy as np

P00 = 1
P01 = 100


@pytest.fixture
def sol():
    kat = finesse.Model()
    kat.parse(
        """
        l L0 P=1
        s s1 L0.p1 ITM.p1

        m ITM R=1 T=0

        sgen sig ITM.mech.z 1 0

        ad a ITM.p1.o 0 n=0 m=0
        ad b ITM.p1.o 0 n=0 m=1

        ad c ITM.p1.o &fsig.f n=0 m=0
        ad d ITM.p1.o &fsig.f n=0 m=1

        ad e ITM.p1.o -&fsig.f n=0 m=0
        ad f ITM.p1.o -&fsig.f n=0 m=1

        gauss g1 L0.p1.o z=0 w0=1e-3

        fsig(1)
        modes(maxtem=1)
        """
    )
    kat.L0.tem(0, 0, P00, 0)
    kat.L0.tem(0, 1, P01, 0)

    return kat.run()


def test_mech_z_scatter_HOM(sol):
    """
    Assert that the correct ratio of input carrier hom get
    scattered into signal sidebands
    """
    assert abs(sol["a"] * np.sqrt(P01 / P00) - sol["b"]) < 1e-15
    assert abs(sol["c"] * np.sqrt(P01 / P00) - sol["d"]) < 1e-15
    assert abs(sol["e"] * np.sqrt(P01 / P00) - sol["f"]) < 1e-15

    assert abs(sol["c"] - sol["e"]) < 1e-15



"""
A components sub-module containing classes for detecting power at
a physical point in a configuration.
"""

import numpy as np
import types
import finesse
import finesse.detectors._pdtypes as pdtypes
from finesse.components.general import Connector, borrows_nodes
from finesse.components.node import NodeDirection, NodeType, Port
from finesse.components.workspace import ConnectorWorkspace
from finesse.parameter import float_parameter
from finesse.detectors.compute.power import PD0Workspace, PD1Workspace, PD2Workspace
from finesse.element import ModelElement

from finesse.detectors.compute.quantum import (
    QShot0Workspace,
    QShotNWorkspace,
)

class ReadoutWorkspace(ConnectorWorkspace):
    pass


class ReadoutOutput(ModelElement):
    def __init__(self, name, readout):
        super().__init__(name)
        self.__readout = readout

    @property
    def readout(self):
        return self.__readout


class Readout(Connector): 
    def __init__(self, name, optical_node, output_detectors=False):
        super().__init__(name)
        self.__output_detectors = output_detectors
        self._add_port("p1", NodeType.OPTICAL)

        if optical_node is not None:
            port = optical_node if isinstance(optical_node, Port) else optical_node.port
            other_node = tuple(o for o in port.nodes if o is not optical_node)[0]

            self.p1._add_node("i", None, optical_node)
            self.p1._add_node("o", None, other_node)
        else:
            self.p1._add_node("i", NodeDirection.INPUT)
            self.p1._add_node("o", NodeDirection.OUTPUT)

    def _get_output_workspaces(self, model):
        return None

    @property
    def optical_node(self):
        if self.p1.i.component != self:
            return self.p1.i

    @property
    def has_mask(self):
        return False

    @property
    def output_detectors(self):
        return self.__output_detectors

    @output_detectors.setter
    def output_detectors(self, value):
        self.__output_detectors = value


@borrows_nodes()
class ReadoutDC(Readout):
    def __init__(self, name, optical_node=None, output_detectors=False):
        super().__init__(name, optical_node, output_detectors=output_detectors)
        
        self._add_port("DC", NodeType.ELECTRICAL)
        self.DC._add_node("o", NodeDirection.OUTPUT)

        self._register_node_coupling("P1i_DC", self.p1.i, self.DC.o)

        self.outputs = types.SimpleNamespace()
        self.outputs.DC = f'{self.name}_DC'

    def _on_add(self, model):
        model.add(ReadoutOutput(self.name+"_DC", self))

    def _get_workspace(self, sim):
        if sim.signal:
            has_DC_node = self.DC.o.full_name in sim.signal.nodes
            
            if not has_DC_node:
                return None # Don't do anything if no nodes included

            ws = ReadoutWorkspace(self, sim, True, True)
            ws.I = np.eye(sim.model_data.num_HOMs, dtype=np.complex128)
            ws.signal.add_fill_function(self._fill_matrix, True)
            ws.frequencies = sim.signal.electrical_frequencies[self.DC.o].frequencies
            return ws
        else:
            return None

    def _get_output_workspaces(self, sim):
        from finesse.detectors.workspace import OutputInformation
        from finesse.detectors.compute.power import pd0_DC_output, PD0Workspace
        wss = []
        for quadrature in ('DC',):
            # Setup a single demodulation photodiode detector for
            # using for outputs
            oinfo = OutputInformation(
                self.name + "_" + quadrature,
                (self.p1.i, ),
                np.float64,
                "W",
                None,
                "W",
                True,
                False
            )
            ws = PD0Workspace(self, sim, oinfo=oinfo)
            ni = sim.carrier.get_node_info(oinfo.nodes[0])
            ws.rhs_index = ni["rhs_index"]
            ws.size = ni["nfreqs"] * ni["nhoms"]
            ws.dc_node_id = sim.carrier.node_id(oinfo.nodes[0])
            ws.set_output_fn(pd0_DC_output)
            wss.append(ws)

        if sim.signal:
            oinfo = OutputInformation(
                self.name + "_shot_noise",
                (self.p1.i, ),
                np.float64,
                "W/rtHz",
                None,
                "ASD",
                True,
                False
            )
            wss.append(
                QShot0Workspace(self, sim, False, output_info=oinfo)
            )

        return wss

    def _fill_matrix(self, ws):
        """
        Computing E.conj() * upper + E * lower.conj()
        """
        for freq in ws.sim.signal.optical_frequencies.frequencies:
            # Get the carrier HOMs for this frequency
            E = ws.sim.carrier.get_out(self.p1.i, freq.audio_carrier_index)
            # is_lower_sb = freq.audio_order < 0
            for efreq in ws.frequencies:
                with ws.sim.signal.component_edge_fill3(
                    ws.owner_id, ws.signal.connections.P1i_DC_idx, freq.index, efreq.index,
                ) as mat:
                    mat[:] = E.conjugate() 


@borrows_nodes()
class ReadoutDCQPD(Readout):
    def __init__(self, name, optical_node=None, output_detectors=False):
        super().__init__(name, optical_node, output_detectors=output_detectors)

        self._add_port("x", NodeType.ELECTRICAL)
        self.x._add_node("o", NodeDirection.OUTPUT)

        self._add_port("y", NodeType.ELECTRICAL)
        self.y._add_node("o", NodeDirection.OUTPUT)

        self._register_node_coupling("P1i_x", self.p1.i, self.x.o)
        self._register_node_coupling("P1i_y", self.p1.i, self.y.o)

        self.outputs = types.SimpleNamespace()
        self.outputs.x = f'{self.name}_X'
        self.outputs.y = f'{self.name}_Y'

    def _on_add(self, model):
        model.add(ReadoutOutput(self.name + "_X", self))
        model.add(ReadoutOutput(self.name + "_Y", self))

    def _get_workspace(self, sim):
        if sim.signal:
            has_x_node = self.x.o.full_name in sim.signal.nodes
            has_y_node = self.y.o.full_name in sim.signal.nodes
            
            if not (has_x_node or has_y_node):
                return None # Don't do anything if no nodes included

            ws = ReadoutWorkspace(self, sim, True, True)
            ws.signal.add_fill_function(self._fill_matrix, True)
            ws.frequencies = sim.signal.electrical_frequencies[self.y.o].frequencies
            ws.Kx = pdtypes.construct_segment_beat_matrix(
                sim.model.mode_index_map,
                pdtypes.XSPLIT
            )
            ws.Ky = pdtypes.construct_segment_beat_matrix(
                sim.model.mode_index_map,
                pdtypes.YSPLIT
            )
            ws.node_car_id = sim.carrier.node_id(self.p1.i)
            return ws
        else:
            return None

    def _get_output_workspaces(self, sim):
        from finesse.detectors.workspace import OutputInformation
        from finesse.detectors.compute.power import pd0_DC_output_segmented, PD0Workspace
        wss = []
        for dof, split in (('X', pdtypes.XSPLIT), ('Y', pdtypes.YSPLIT)):
            # Setup a single demodulation photodiode detector for
            # using for outputs
            oinfo = OutputInformation(
                self.name + "_" + dof,
                (self.p1.i, ),
                np.float64,
                "W",
                None,
                "W",
                True,
                False
            )
            ws = PD0Workspace(self, sim, oinfo=oinfo)
            ws.dc_node_id = sim.carrier.node_id(oinfo.nodes[0])
            ws.tmp = np.zeros(ws.size, dtype=complex)
            ws.K = pdtypes.construct_segment_beat_matrix(
                sim.model.mode_index_map, split
            )
            ws.set_output_fn(pd0_DC_output_segmented)
            wss.append(ws)

        return wss

    def _fill_matrix(self, ws):
        for freq in ws.sim.signal.optical_frequencies.frequencies:
            # Get the carrier HOMs for this frequency
            cidx = freq.audio_carrier_index
            rhs_idx = ws.sim.carrier.field(self.p1.i, cidx, 0)
            Ec = np.conjugate(ws.sim.carrier.out_view[rhs_idx:(rhs_idx + ws.sim.model_data.num_HOMs)])
            # is_lower_sb = freq.audio_order < 0
            for efreq in ws.frequencies:
                if ws.signal.connections.P1i_x_idx > -1:
                    with ws.sim.signal.component_edge_fill3(
                        ws.owner_id, ws.signal.connections.P1i_x_idx, freq.index, efreq.index,
                    ) as mat:
                        mat[:] = np.dot(ws.Kx, Ec)

                if ws.signal.connections.P1i_y_idx > -1:
                    with ws.sim.signal.component_edge_fill3(
                        ws.owner_id, ws.signal.connections.P1i_y_idx, freq.index, efreq.index,
                    ) as mat:
                        mat[:] = np.dot(ws.Ky, Ec)


@borrows_nodes()
@float_parameter("f", "Frequency")
@float_parameter("phase", "Phase")
class ReadoutRF(Readout):
    def __init__(self, name, optical_node=None, *, f=None, phase=0, output_detectors=False):
        super().__init__(name, optical_node, output_detectors=output_detectors)

        self.f = f
        self.phase = phase

        self._add_port("I", NodeType.ELECTRICAL)
        self.I._add_node("o", NodeDirection.OUTPUT)
        self._add_port("Q", NodeType.ELECTRICAL)
        self.Q._add_node("o", NodeDirection.OUTPUT)

        self._register_node_coupling("P1i_I", self.p1.i, self.I.o)
        self._register_node_coupling("P1i_Q", self.p1.i, self.Q.o)

        self.outputs = types.SimpleNamespace()
        self.outputs.I = f'{self.name}_I'
        self.outputs.Q = f'{self.name}_Q'

    @property
    def optical_node(self):
        if self.p1.i.component != self:
            return self.p1.i

    def _on_add(self, model):
        model.add(ReadoutOutput(self.name+"_I", self))
        model.add(ReadoutOutput(self.name+"_Q", self))

    def _get_workspace(self, sim):
        if sim.signal:
            has_I_node = self.I.o.full_name in sim.signal.nodes
            has_Q_node = self.Q.o.full_name in sim.signal.nodes

            if not (has_I_node or has_Q_node):
                return None # Don't do anything if no nodes included

            ws = ReadoutWorkspace(self, sim, True, True)
            ws.signal.add_fill_function(self._fill_matrix, True)
            ws.frequencies = sim.signal.electrical_frequencies[self.I.o if has_I_node else self.Q.o].frequencies
            ws.dc_node_id = sim.carrier.node_id(self.p1.i)
            return ws
        else:
            return None

    def _get_output_workspaces(self, sim):
        from finesse.detectors.workspace import OutputInformation
        from finesse.detectors.compute.power import pd1_DC_output
        wss = []
        for quadrature in ('I', 'Q'):
            # Setup a single demodulation photodiode detector for
            # using for outputs
            oinfo = OutputInformation(
                self.name + "_" + quadrature,
                (self.p1.i, ),
                np.float64,
                "W",
                None,
                "W",
                True,
                False
            )
            poff = 90 if quadrature == 'Q' else 0
            ws = PD1Workspace(self, sim, phase_offset=poff, oinfo=oinfo)
            ws.is_f_changing = self.f.is_changing
            ws.is_phase_changing = self.phase.is_changing
            ws.output_real = True
            ws.dc_node_id = sim.carrier.node_id(oinfo.nodes[0])
            ws.homs = ws.sim.model_data.homs_view
            ws.set_output_fn(pd1_DC_output)

            if not ws.is_f_changing:
                # precompute these things if nothing is changing
                ws.update_parameter_values()
                ws.update_beats()

            wss.append(ws)

        if sim.signal:
            oinfo = OutputInformation(
                self.name + "_shot_noise",
                (self.p1.i, ),
                np.float64,
                "W/rtHz",
                None,
                "ASD",
                True,
                False
            )
            wss.append(
                QShotNWorkspace(self, sim, 1, False, output_info=oinfo)
            )
        return wss

    def _fill_matrix(self, ws):
        # 1/2 from demod gain
        factorI = (
            -1 # ddb - can't see why we need this minus sign.
            # Makes us match up with pd2 in v2 and v3 though
            * ws.sim.model_data.EPSILON0_C
            * np.exp(-1j * ws.values.phase * finesse.constants.DEG2RAD)
        )
        cfactorI = factorI.conjugate()
        factorQ = 1j * factorI
        cfactorQ = factorQ.conjugate()

        # Need to zero everything first
        for f1 in ws.sim.carrier.optical_frequencies.frequencies:
            for f2 in ws.sim.carrier.optical_frequencies.frequencies:
                df = f1.f - f2.f
                if df == ws.values.f or df == -ws.values.f:
                    if ws.signal.connections.P1i_I_idx >= 0:
                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_I_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f2.index)["audio_lower_index"],
                            0,
                        ) as mat:
                            mat[:] = 0
                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_I_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f1.index)["audio_upper_index"],
                            0,
                        ) as mat:
                            mat[:] = 0

                    if ws.signal.connections.P1i_Q_idx >= 0:
                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_Q_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f2.index)["audio_lower_index"],
                            0,
                        ) as mat:
                            mat[:] = 0

                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_Q_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f1.index)["audio_upper_index"],
                            0,
                        ) as mat:
                            mat[:] = 0

        for f1 in ws.sim.carrier.optical_frequencies.frequencies:
            for f2 in ws.sim.carrier.optical_frequencies.frequencies:
                df = f1.f - f2.f
                # Get the carrier HOMs for this frequency
                rhs_idx = ws.sim.carrier.field(self.p1.i, f1.index, 0)
                E1 = ws.sim.carrier.out_view[rhs_idx:(rhs_idx + ws.sim.model_data.num_HOMs)]
                rhs_idx = ws.sim.carrier.field(self.p1.i, f2.index, 0)
                E2c = np.conjugate(ws.sim.carrier.out_view[rhs_idx:(rhs_idx + ws.sim.model_data.num_HOMs)])
                
                if df == -ws.values.f:
                    if ws.signal.connections.P1i_I_idx >= 0:
                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_I_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f2.index)["audio_lower_index"],
                            0,
                        ) as mat:
                            # This will apply a conjugation internally as it's
                            # a lower SB connection
                            mat[:] += factorI.conjugate() * E1

                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_I_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f1.index)["audio_upper_index"],
                            0,
                        ) as mat:
                            mat[:] += factorI.conjugate() * E2c

                    if ws.signal.connections.P1i_Q_idx >= 0:
                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_Q_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f2.index)["audio_lower_index"],
                            0,
                        ) as mat:
                            # This will apply a conjugation internally as it's 
                            # a lower SB connection
                            mat[:] += factorQ.conjugate() * E1

                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_Q_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f1.index)["audio_upper_index"],
                            0,
                        ) as mat:
                            mat[:] += factorQ.conjugate() * E2c

                if df == ws.values.f:
                    if ws.signal.connections.P1i_I_idx >= 0:
                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_I_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f2.index)["audio_lower_index"],
                            0,
                        ) as mat:
                            mat[:] += factorI * E1

                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_I_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f1.index)["audio_upper_index"],
                            0,
                        ) as mat:
                            mat[:] += factorI * E2c

                    if ws.signal.connections.P1i_Q_idx >= 0:
                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_Q_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f2.index)["audio_lower_index"],
                            0,
                        ) as mat:
                            mat[:] += factorQ * E1

                        with ws.sim.signal.component_edge_fill3(
                            ws.owner_id,
                            ws.signal.connections.P1i_Q_idx,
                            ws.sim.carrier.optical_frequencies.get_info(f1.index)["audio_upper_index"],
                            0,
                        ) as mat:
                            mat[:] += factorQ * E2c



"""Dielectric interface type components representing physical mirrors."""

import logging
import types

import numpy as np

from finesse.parameter import float_parameter, bool_parameter
from finesse.utilities import refractive_index

from finesse.components.general import InteractionType, NoiseType, DOFDefinition
from finesse.components.surface import Surface
from finesse.components.node import NodeDirection, NodeType

LOGGER = logging.getLogger(__name__)


@float_parameter("R", "Reflectivity", validate="_check_R", setter="set_RTL")
@float_parameter("T", "Transmission", validate="_check_T", setter="set_RTL")
@float_parameter("L", "Loss", validate="_check_L", setter="set_RTL")
@float_parameter(
    "phi", "Microscopic tuning (360 degrees = 1 default wavelength)", units="degrees"
)
@float_parameter(
    "Rcx",
    "Radius of curvature (x)",
    units="m",
    validate="_check_Rc",
    is_geometric=True,
)
@float_parameter(
    "Rcy",
    "Radius of curvature (y)",
    units="m",
    validate="_check_Rc",
    is_geometric=True,
)
@float_parameter("xbeta", "Yaw misalignment", units="radians")
@float_parameter("ybeta", "Pitch misalignment", units="radians")
@bool_parameter("misaligned", "Misaligns mirror reflection (R=0 when True)")
class Mirror(Surface):
    """The mirror component represents a thin dielectric surface with associated
    properties such as reflectivity, tuning, and radius of curvature. It has two optical
    ports `p1` and `p2` which decsribe the two beams incident on either side of this
    surface. It also has a mechanical port `mech` which has nodes for longitudinal, yaw,
    and pitch motions. These mechanical nodes are purely for exciting small signal
    oscillations of the mirror. Static offsets in longitudinal displacements are set by
    the `phi` parameter (in units of degrees), yaw by the `xbeta` parameter, and pitch
    the `ybeta` parameter.

    Parameters
    ----------
    name : str
        Name of newly created mirror.

    R : float, optional
        Reflectivity of the mirror; defaults to 0.5.

    T : float, optional
        Transmittance of the mirror; defaults to 0.5.

    L : float, optional
        Loss of the mirror; defaults to 0.0.

    phi : float, optional
        Tuning of the mirror (in degrees); defaults to 0.0.

    Rc : float or container of two floats, optional
        The radius of curvature of the mirror (in metres); defaults to a flat mirror (`Rc=np.inf`).
        Astigmatic mirrors can also be set with `Rc=(Rcx, Rcy)`.

    xbeta, ybeta : float, optional
        Misalignment of the mirror in yaw and pitch in units of radians
    """

    def __init__(
        self,
        name,
        R=None,
        T=None,
        L=None,
        phi=0,
        Rc=np.inf,
        xbeta=0,
        ybeta=0,
        misaligned=False,
    ):
        super().__init__(name, R, T, L, phi, Rc, xbeta, ybeta)
        self.misaligned = misaligned

        self._add_port("p1", NodeType.OPTICAL)
        self.p1._add_node("i", NodeDirection.INPUT)
        self.p1._add_node("o", NodeDirection.OUTPUT)

        self._add_port("p2", NodeType.OPTICAL)
        self.p2._add_node("i", NodeDirection.INPUT)
        self.p2._add_node("o", NodeDirection.OUTPUT)

        # Optic to optic couplings
        self._register_node_coupling(
            "P1i_P1o",
            self.p1.i,
            self.p1.o,
            interaction_type=InteractionType.REFLECTION,
            # enabled_check=lambda: self.R > 0 and not self.R.is_changing
        )
        self._register_node_coupling(
            "P2i_P2o",
            self.p2.i,
            self.p2.o,
            interaction_type=InteractionType.REFLECTION,
            # enabled_check=lambda: self.R > 0 and not self.R.is_changing
        )
        self._register_node_coupling(
            "P1i_P2o",
            self.p1.i,
            self.p2.o,
            interaction_type=InteractionType.TRANSMISSION,
            # enabled_check=lambda: self.T > 0 and not self.T.is_changing
        )
        self._register_node_coupling(
            "P2i_P1o",
            self.p2.i,
            self.p1.o,
            interaction_type=InteractionType.TRANSMISSION,
            # enabled_check=lambda: self.T > 0 and not self.T.is_changing
        )

        # Mirror motion couplings
        self._add_port("mech", NodeType.MECHANICAL)
        self.mech._add_node("z", NodeDirection.OUTPUT)
        self.mech._add_node("yaw", NodeDirection.OUTPUT)
        self.mech._add_node("pitch", NodeDirection.OUTPUT)
        self.mech._add_node("F_z", NodeDirection.INPUT)
        self.mech._add_node("F_yaw", NodeDirection.INPUT)
        self.mech._add_node("F_pitch", NodeDirection.INPUT)

        # Optic to motion couplings
        self._register_node_coupling("P1i_Fz", self.p1.i, self.mech.F_z)
        self._register_node_coupling("P2i_Fz", self.p2.i, self.mech.F_z)
        self._register_node_coupling("P1o_Fz", self.p1.o, self.mech.F_z)
        self._register_node_coupling("P2o_Fz", self.p2.o, self.mech.F_z)
        self._register_node_coupling("P1i_Fyaw", self.p1.i, self.mech.F_yaw)
        self._register_node_coupling("P2i_Fyaw", self.p2.i, self.mech.F_yaw)
        self._register_node_coupling("P1o_Fyaw", self.p1.o, self.mech.F_yaw)
        self._register_node_coupling("P2o_Fyaw", self.p2.o, self.mech.F_yaw)
        self._register_node_coupling("P1i_Fpitch", self.p1.i, self.mech.F_pitch)
        self._register_node_coupling("P2i_Fpitch", self.p2.i, self.mech.F_pitch)
        self._register_node_coupling("P1o_Fpitch", self.p1.o, self.mech.F_pitch)
        self._register_node_coupling("P2o_Fpitch", self.p2.o, self.mech.F_pitch)

        # motion to optic coupling: phase coupling on reflection
        self._register_node_coupling("Z_P1o", self.mech.z, self.p1.o)
        self._register_node_coupling("Z_P2o", self.mech.z, self.p2.o)
        self._register_node_coupling("yaw_P1o", self.mech.yaw, self.p1.o)
        self._register_node_coupling("yaw_P2o", self.mech.yaw, self.p2.o)
        self._register_node_coupling("pitch_P1o", self.mech.pitch, self.p1.o)
        self._register_node_coupling("pitch_P2o", self.mech.pitch, self.p2.o)

        self.__changing_check = set(
            (self.R, self.T, self.L, self.phi, self.xbeta, self.ybeta)
        )

        # Define typical degrees of freedom for this component
        self.dofs = types.SimpleNamespace()
        self.dofs.z = DOFDefinition(self.phi, self.mech.z, 1)
        self.dofs.yaw = DOFDefinition(self.xbeta, self.mech.yaw, 1)
        self.dofs.pitch = DOFDefinition(self.ybeta, self.mech.pitch, 1)

    def _resymbolise_ABCDs(self):
        # -> reflections
        self.__symbolise_ABCD(self.p1.i, self.p1.o, "x")
        self.__symbolise_ABCD(self.p1.i, self.p1.o, "y")
        self.__symbolise_ABCD(self.p2.i, self.p2.o, "x")
        self.__symbolise_ABCD(self.p2.i, self.p2.o, "y")
        # -> transmissions
        self.__symbolise_ABCD(self.p1.i, self.p2.o, "x")
        self.__symbolise_ABCD(self.p1.i, self.p2.o, "y")
        self.__symbolise_ABCD(self.p2.i, self.p1.o, "x")
        self.__symbolise_ABCD(self.p2.i, self.p1.o, "y")

    def __symbolise_ABCD(self, from_node, to_node, direction):
        if direction == "x":
            Rc = self.Rcx.ref
        else:
            Rc = self.Rcy.ref

        # reflection
        if self.interaction_type(from_node, to_node) == InteractionType.REFLECTION:
            nr = refractive_index(from_node, symbolic=True)

            C = 2 * nr / Rc
            if from_node.port is self.p1:
                C *= -1

        # transmission
        else:
            nr1 = refractive_index(from_node, symbolic=True)
            nr2 = refractive_index(to_node, symbolic=True)

            C = (nr2 - nr1) / Rc
            if from_node.port is not self.p1:
                C *= -1

        M_sym = np.array([[1.0, 0.0], [C, 1.0]])

        key = (from_node, to_node, direction)
        # For mirrors the symbol nr can change when connected up to spaces
        # in a model so here we update the ABCD matrix entry
        if key in self._abcd_matrices:
            self._abcd_matrices[key] = M_sym, np.array(M_sym, dtype=np.float64)
        # Otherwise just register the new ABCD matrix as usual
        else:
            self.register_abcd_matrix(M_sym, (from_node, to_node, direction))

    @property
    def abcd11x(self):
        """Numeric ABCD matrix from port 1 to port 1 in the tangential plane.

        Equivalent to ``mirror.ABCD(1, 1, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(1, 1, "x")

    @property
    def abcd11y(self):
        """Numeric ABCD matrix from port 1 to port 1 in the sagittal plane.

        Equivalent to ``mirror.ABCD(1, 1, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(1, 1, "y")

    @property
    def abcd22x(self):
        """Numeric ABCD matrix from port 2 to port 2 in the tangential plane.

        Equivalent to ``mirror.ABCD(2, 2, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(2, 2, "x")

    @property
    def abcd22y(self):
        """Numeric ABCD matrix from port 2 to port 2 in the sagittal plane.

        Equivalent to ``mirror.ABCD(2, 2, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(2, 2, "y")

    @property
    def abcd12x(self):
        """Numeric ABCD matrix from port 1 to port 2 in the tangential plane.

        Equivalent to ``mirror.ABCD(1, 2, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(1, 2, "x")

    @property
    def abcd12y(self):
        """Numeric ABCD matrix from port 1 to port 2 in the sagittal plane.

        Equivalent to ``mirror.ABCD(1, 2, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(1, 2, "y")

    @property
    def abcd21x(self):
        """Numeric ABCD matrix from port 2 to port 1 in the tangential plane.

        Equivalent to ``mirror.ABCD(2, 1, "x")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(2, 1, "x")

    @property
    def abcd21y(self):
        """Numeric ABCD matrix from port 2 to port 1 in the sagittal plane.

        Equivalent to ``mirror.ABCD(2, 1, "y")``.

        :getter: Returns a copy of the (numeric) ABCD matrix for this coupling (read-only).
        """
        return self.ABCD(2, 1, "y")

    def ABCD(
        self,
        from_node,
        to_node,
        direction="x",
        symbolic=False,
        copy=True,
        retboth=False,
    ):
        r"""Returns the ABCD matrix of the mirror for the specified coupling.

        In both cases below, the sign of the radius is defined such that :math:`R_c`
        is negative if the centre of the sphere is located in the direction of propagation.

        .. rubric:: Transmission

        .. _fig_abcd_mirror_transmission:
        .. figure:: ../images/abcd_mi.*
            :align: center

        For transmission this is given by,

        .. math::
            M_{t} = \begin{pmatrix}
                        1 & 0 \\
                        \frac{n_2 - n_1}{R_c} & 1
                    \end{pmatrix},

        where :math:`n_2` and :math:`n_1` are the indices of refraction of the spaces
        connected to the mirror and :math:`R_c` is the radius of curvature of the mirror.

        The matrix for transmission in the opposite direction of propagation is identical.

        .. rubric:: Reflection

        .. _fig_abcd_mirror_reflection:
        .. figure:: ../images/abcd_mr.*
            :align: center

        In the case of reflection the matrix is,

        .. math::
            M_{r} = \begin{pmatrix}
                        1 & 0 \\
                        -\frac{2n_1}{R_c} & 1
                    \end{pmatrix}.

        Reflection at the back surface can be described by the same type of matrix by setting
        the :math:`C` element to :math:`C = 2n_2/R_c`.

        See :meth:`.Connector.ABCD` for descriptions of parameters, return values and possible
        exceptions.
        """
        return super().ABCD(from_node, to_node, direction, symbolic, copy, retboth)

    def _get_workspace(self, sim):
        from finesse.simulations.basematrix import CarrierSignalMatrixSimulation

        if isinstance(sim, CarrierSignalMatrixSimulation):
            from finesse.components.modal.mirror import (
                mirror_carrier_fill,
                mirror_signal_fill,
                mirror_fill_qnoise,
                MirrorWorkspace,
            )

            _, is_changing = self._eval_parameters()

            misaligned = self.xbeta.value != 0 or self.ybeta.value != 0
            refill = (
                sim.is_component_in_mismatch_couplings(self)
                or (misaligned and sim.trace_forest.contains_comp(self))
                or sim.carrier.any_frequencies_changing
                or (sim.signal.any_frequencies_changing if sim.signal else False)
                or len(is_changing)
            )

            ws = MirrorWorkspace(self, sim, refill)
            # This assumes that nr1/nr2 cannot change during a simulation
            ws.nr1 = refractive_index(self.p1)
            ws.nr2 = refractive_index(self.p2)

            ws.carrier.add_fill_function(mirror_carrier_fill, refill)
            ws.signal.add_fill_function(mirror_signal_fill, refill)

            # Initialise the ABCD matrix memory-views
            if sim.is_modal:
                ws.abcd_p1p1_x = self.ABCD(self.p1.i, self.p1.o, "x", copy=False)
                ws.abcd_p1p1_y = self.ABCD(self.p1.i, self.p1.o, "y", copy=False)
                ws.abcd_p2p2_x = self.ABCD(self.p2.i, self.p2.o, "x", copy=False)
                ws.abcd_p2p2_y = self.ABCD(self.p2.i, self.p2.o, "y", copy=False)

                ws.abcd_p1p2_x = self.ABCD(self.p1.i, self.p2.o, "x", copy=False)
                ws.abcd_p1p2_y = self.ABCD(self.p1.i, self.p2.o, "y", copy=False)
                ws.abcd_p2p1_x = self.ABCD(self.p2.i, self.p1.o, "x", copy=False)
                ws.abcd_p2p1_y = self.ABCD(self.p2.i, self.p1.o, "y", copy=False)

                ws.set_knm_info(
                    "P1i_P1o",
                    abcd_x=ws.abcd_p1p1_x,
                    abcd_y=ws.abcd_p1p1_y,
                    nr_from=ws.nr1,
                    nr_to=ws.nr1,
                    is_transmission=False,
                    beta_x=self.xbeta,
                    beta_x_factor=2,
                    beta_y=self.ybeta,
                    beta_y_factor=-2,
                )
                ws.set_knm_info(
                    "P2i_P2o",
                    abcd_x=ws.abcd_p2p2_x,
                    abcd_y=ws.abcd_p2p2_y,
                    nr_from=ws.nr2,
                    nr_to=ws.nr2,
                    is_transmission=False,
                    beta_x=self.xbeta,
                    beta_x_factor=2,
                    beta_y=self.ybeta,
                    beta_y_factor=2,
                )
                ws.set_knm_info(
                    "P1i_P2o",
                    abcd_x=ws.abcd_p1p2_x,
                    abcd_y=ws.abcd_p1p2_y,
                    nr_from=ws.nr1,
                    nr_to=ws.nr2,
                    is_transmission=True,
                    beta_x=self.xbeta,
                    beta_x_factor=-(1 - ws.nr1 / ws.nr2),
                    beta_y=self.ybeta,
                    beta_y_factor=(1 - ws.nr1 / ws.nr2),
                )
                ws.set_knm_info(
                    "P2i_P1o",
                    abcd_x=ws.abcd_p2p1_x,
                    abcd_y=ws.abcd_p2p1_y,
                    nr_from=ws.nr2,
                    nr_to=ws.nr1,
                    is_transmission=True,
                    beta_x=self.xbeta,
                    beta_x_factor=-(1 - ws.nr2 / ws.nr1),
                    beta_y=self.ybeta,
                    beta_y_factor=-(1 - ws.nr2 / ws.nr1),
                )

            if sim.signal:
                ws.signal.set_fill_noise_function(NoiseType.QUANTUM, mirror_fill_qnoise)
            return ws



"""Useful common utility functions and classes used throughout the Finesse package.

Listed below are all the sub-modules of the ``utilities`` module with a brief
description of the contents of each.
"""

# TODO: sort out which things get imported at the module level here
from .components import refractive_index
from .homs import make_modes, insert_modes
from .text import (
    ngettext,
    option_list,
    format_section,
    format_bullet_list,
    add_linenos,
    stringify,
    stringify_graph_gml,
)
from finesse.utilities.misc import (
    check_name,
    pairwise,
    valid_name,
    is_iterable,
    changed_params,
    opened_file,
    logs,
    graph_layouts,
    networkx_layouts,
    graphviz_layouts,
)
from finesse.utilities.units import SI, SI_LABEL, SI_VALUE

__all__ = (
    "refractive_index",
    "make_modes",
    "insert_modes",
    "ngettext",
    "option_list",
    "format_section",
    "format_bullet_list",
    "add_linenos",
    "stringify",
    "stringify_graph_gml",
    "check_name",
    "pairwise",
    "valid_name",
    "is_iterable",
    "changed_params",
    "opened_file",
    "logs",
    "graph_layouts",
    "networkx_layouts",
    "graphviz_layouts",
    "SI",
    "SI_LABEL",
    "SI_VALUE",
)



"""Tests for making sure that order in katscript doesn't matter."""

from finesse.script import parse


def test_space_instruction_order_doesnt_matter():
    """Test that the order in which a space is defined in kat script does not matter."""
    # Shouldn't throw an error.
    parse(
        """
        l L0 P=1

        s s0 L0.p1 ITM.p1

        s CAV ITM.p2 ETM.p1 L=1  # Defined before the ITM and ETM components.

        m ITM R=0.99 T=0.01 Rc=-10
        m ETM R=0.99 T=0.01 Rc=10

        modes(even, 4)
        """
    )


def test_cavity_property_detector():
    """Cavity property detector should parse after cavity.

    The cavity property detector takes a cavity as a parameter, but cavities are `build_last`
    components so get built after everything that isn't. Cavity property detectors therefore have
    to be built after cavities.
    """
    # Shouldn't throw an error.
    parse(
        """
        l L0 P=1
        s s0 L0.p1 ITM.p1

        m ITM Rc=-2
        s sc ITM.p2 ETM.p1 L=1
        m ETM Rc=2

        cav FP ITM.p2
        cp gx FP g x

        modes(maxtem=2)

        xaxis(ITM.phi, lin, 0, 180, 100)
        """
    )


def test_gouy_detector():
    """Gouy detector should parse in the second build step.

    The gouy detector has implicit dependencies like `cavity`, in that it requires for there to
    exist a path between its start and end nodes by the time it itself gets built. It therefore is
    set to `build_last` in the :class:`.KatSpec`. This test checks that the gouy detector still
    parses correctly.
    """
    # Shouldn't throw an error.
    parse(
        """
        l L0 P=1
        s s0 L0.p1 ITM.p1

        m ITM Rc=-2
        s sc ITM.p2 ETM.p1 L=1
        m ETM Rc=2

        cav FP ITM.p2
        gouy gFP from_node=ITM.p2 to_node=ETM.p1

        modes(maxtem=2)
        xaxis(ITM.phi, lin, 0, 180, 100)
        """
    )



"""Tasks involving tracing beams and computing ABCD matrices."""

import logging
from .actions import ABCD, BeamTrace, PropagateBeam, PropagateAstigmaticBeam

LOGGER = logging.getLogger(__name__)


def abcd(model, **kwargs):
    """Executes a :class:`.ABCD` analysis on a model.

    Other Parameters
    ----------------
    **kwargs
        Arguments to pass to the ABCD routine. See :meth:`~.Model.ABCD` for the available options.

    Returns
    -------
    out : :class:`.ABCDSolution`
        The solution object for the ABCD matrix calculation.
    """
    analysis = ABCD(**kwargs)
    return analysis.run(model)


def beam_trace(model, **kwargs):
    """Executes a :class:`.BeamTrace` analysis on a model.

    Other Parameters
    ----------------
    **kwargs
        Arguments to pass to the beam trace routine. See :meth:`.Model.beam_trace` for the
        available options.

    Returns
    -------
    out : :class:`.BeamTraceSolution`
        The solution object for the model beam trace.
    """
    analysis = BeamTrace(**kwargs)
    return analysis.run(model)


def propagate_beam(model, **kwargs):
    """Executes a :class:`.PropagateBeam` analysis on a model.

    Other Parameters
    ----------------
    **kwargs
        Arguments to pass to the routine. See :meth:`.Model.propagate_beam` for the
        available options.

    Returns
    -------
    out : :class:`.BeamTraceSolution`
        The solution object for the single plane beam propagation.
    """
    analysis = PropagateBeam(**kwargs)
    return analysis.run(model)


def propagate_beam_astig(model, **kwargs):
    """Executes a :class:`.PropagateAstigmaticBeam` analysis on a model.

    Other Parameters
    ----------------
    **kwargs
        Arguments to pass to the routine. See :meth:`.Model.propagate_beam_astig` for the
        available options.

    Returns
    -------
    out : :class:`.BeamTraceSolution`
        The solution object for the astigmatic beam propagation.
    """
    analysis = PropagateAstigmaticBeam(**kwargs)
    return analysis.run(model)



import finesse
import numpy as np

from finesse.exceptions import ParameterLocked
from finesse.element import ModelElement
from finesse.parameter import float_parameter
from finesse.solutions.base import BaseSolution
from finesse.analysis.actions import Action
from finesse.analysis.runners import run_fsig_sweep


@float_parameter("ASD", "Amplitude spectral density", units='ASD')
class ClassicalNoise(ModelElement):
    def __init__(self, name, node, ASD):
        super().__init__(name)
        self._add_to_model_namespace = True
        self._namespace = ".noises"
        self.node = node
        self.ASD = ASD


class NoiseSolution(BaseSolution):
    def __init__(self, name, f, output_nodes, noises, parents=None):
        super().__init__(name, parents)
        self.f = f
        self.dtype = np.dtype([(n.name, float) for n in noises])
        self.noises = {
            onode: np.zeros(f.shape, dtype=self.dtype) for onode in output_nodes
        }

    def total_ASD(self, output_node):
        return np.sqrt(sum(v ** 2 for v in self.noises[output_node].values()))

    def plot(
        self, output_node, noises=None, show_total=True, figsize_scale=1, ax=None,
    ):
        import matplotlib.pyplot as plt

        if ax is None:
            fig, ax = plt.subplots()
        if not isinstance(output_node, str):
            output_node = output_node.full_name

        if output_node not in self.noises:
            raise Exception(
                f"The noise at {output_node} was not calculated, please recalculate."
            )

        for k, v in self.noises[output_node].items():
            if noises is not None and k not in noises:
                continue
            if hasattr(v, "shape"):
                ax.loglog(self.f, v, label=k)
            else:
                ax.loglog(self.f, v * np.ones_like(self.f), label=k)

        if (noises is not None and len(noises) > 1) or (show_total and noises is None):
            ax.loglog(
                self.f, self.total_ASD(output_node), c="k", ls=":", lw=3, label="Total"
            )

        ax.legend()
        ax.set_title(f"Noise at {output_node}")
        ax.set_xlabel("Frequency [Hz]")
        ax.set_ylabel("ASD [UNIT/rtHz]")
        plt.gcf().tight_layout()
        return ax


def get_loop_network(model):
    import networkx
    import sympy
    from finesse.components import NodeType, Wire

    net = networkx.digraph.DiGraph()

    def is_electric(x):
        return all(p.type == NodeType.ELECTRICAL for p in x.ports)

    optics_in = set()  # nodes going into the optics
    optics_out = set()  # nodes coming out of the optics
    Omega = sympy.var("Omega")
    noise_nodes = tuple(n.node.full_name for n in model.noises)

    for i, o, d in model.network.edges(data=True):
        # Keep any edge going into an electronic element
        if (
            d["out_ref"]().type == NodeType.ELECTRICAL
            or d["in_ref"]().type == NodeType.ELECTRICAL
        ):
            if d["in_ref"]().connection or i in noise_nodes or o in noise_nodes:
                # Store the nodes that inject into an optical node
                # and those optical nodes that feed into the electronics
                if d["in_ref"]().type == NodeType.OPTICAL:
                    optics_out.add(d["out_ref"]())
                elif d["out_ref"]().type == NodeType.OPTICAL:
                    optics_in.add(d["in_ref"]())
                else:
                    el = d["owner"]()
                    sym = sympy.var(el.name)
                    if type(el) is Wire:
                        if not el.delay.is_changing:
                            sym = sympy.exp(-sympy.I * Omega * float(el.delay))
                        elif el.delay == 0:
                            sym = 1
                    net.add_edge(i, o, symbol=sym)
        elif d["in_ref"]().type == NodeType.MECHANICAL and i in noise_nodes:
            # Mechanical noise inputs need to be included so we can project them
            # These are really just inputs into the optical system.
            # I guess we could also have mechanical to electrical sensors...
            optics_in.add(d["in_ref"]())

    optic_couplings = {}
    # Add in placeholds for any opto-mechanic coupling
    for i in optics_in:
        for o in optics_out:
            sym = sympy.var(
                f"{i.full_name.replace('.','')}__{o.full_name.replace('.','')}"
            )
            net.add_edge(i.full_name, o.full_name, symbol=sym)
            optic_couplings[sym] = (i, o)
    return net, optic_couplings


def nx_to_coo_sparse(G, nodelist=None):
    from itertools import chain

    if nodelist is None:
        nodelist = list(G)
    nlen = len(nodelist)

    if nlen == 0:
        raise Exception("Graph has no nodes or edges")

    if len(nodelist) != len(set(nodelist)):
        raise Exception("Ambiguous ordering: `nodelist` contained duplicates.")

    index = dict(zip(nodelist, range(nlen)))
    # build I-M sparse matrix
    coefficients = zip(
        *chain(
            # Add in I matrix
            ((i, i, 1) for i in range(nlen)),
            # negative off-diagonal elements
            (
                (index[u], index[v], -d["symbol"])
                for u, v, d in G.edges(nodelist, data=True)
                if u in index and v in index
            ),
        )
    )

    try:
        row, col, data = coefficients
    except ValueError:
        # there is no edge in the subgraph
        row, col, data = [], [], []

    return row, col, data



class NoiseAnalysis(Action):
    def __init__(
        self,
        name,
        mode,
        start,
        stop,
        steps,
        output_node,
        *opt_output_nodes,
        noises=None,
    ):
        super().__init__(name)
        if mode not in ("lin", "log"):
            raise Exception(f"mode must be lin or log")

        if mode == "lin":
            self.f = np.linspace(start, stop, steps)
        else:
            self.f = np.geomspace(start, stop, steps)

        if noises is None:
            self.noises = tuple(n.name for n in output_node.component._model.noises)
        else:
            self.noises = tuple(n.name for n in noises)

        self.output_nodes = tuple(
            (output_node.full_name, *(o.full_name for o in opt_output_nodes))
        )
        self._info.parameters_changing = tuple(("fsig.f",),)
        self._info.makes_solution = True

    def fill_info(self, p_info):
        info = self.copy_info()
        p_info.add(info)

    def setup(self, s_prev, model: finesse.model.Model):
        import sympy
        from sympy.matrices import SparseMatrix
        from sympy import Matrix

        ws = NoiseAnalysisWorkspace(s_prev, model)
        ws.s_prev = s_prev
        ws.model = model
        ws.info = self.copy_info()
        ws.fn_do = self.do
        ws.params = tuple(
            finesse.analysis.actions.get_param(model, p)
            for p in self._info.parameters_changing
        )
        for p in ws.params:
            if not p.is_tunable:
                raise ParameterLocked(
                    f"{repr(p)} must set as tunable " "before building the simulation"
                )
        # Setup the symbolic sparse matrix for solving noise propagations
        ws.loop_network, oc = get_loop_network(model)
        ws.noises = tuple(n for n in model.noises if n.name in self.noises)
        assert len(ws.noises) > 0
        # returns a I-M matrix for solving
        rcv = nx_to_coo_sparse(ws.loop_network)
        ws.M_nodes = nodes = tuple(ws.loop_network.nodes)
        M = SparseMatrix(
            None, {k: v for k, v in zip(tuple(zip(rcv[0], rcv[1])), rcv[2])}
        )
        # Always ensure that the Signal frequency variable
        # is present as we always fill that
        ws.Omega = sympy.var("Omega")
        ws.syms = M.free_symbols | set((ws.Omega,))
        ws.Minv = Matrix(M).inv()
        # Fill up a dict of how each noise
        # couples into each output node
        ws.TF = tuple([] for i in range(len(self.output_nodes)))
        for i, onode in enumerate(self.output_nodes):
            for j, noise in enumerate(ws.noises):
                noise_tf = ws.Minv[
                    nodes.index(noise.node.full_name), nodes.index(onode)
                ].expand()
                ws.TF[i].append(
                    sympy.lambdify(ws.syms, abs(noise_tf) ** 2, "numpy", dummify=False)
                )
        ws.values = {s: 0 for s in ws.syms}
        ws.OMEGA = ws.values[ws.Omega] = 2 * np.pi * self.f
        ws.lambdas = tuple(noise.ASD.lambdify(ws.model.fsig.f) for noise in ws.noises)
        # Get the simulations to do some calculations with in the do step
        ws.carrier = model.carrier_simulation
        try:
            ws.signal = model.signal_simulation
            ws.optical_couplings = {}
            ins = []
            outs = []
            # Store optical coulpings node indices for injecting later
            for key, (inode, onode) in oc.items():
                idx = ws.signal.field(inode, 0, 0)
                odx = ws.signal.field(onode, 0, 0)
                if idx in ins:
                    x = ins.index(idx)
                else:
                    ins.append(idx)
                    x = len(ins) - 1
                if odx in outs:
                    y = outs.index(odx)
                else:
                    outs.append(odx)
                    y = len(outs) - 1
                ws.optical_couplings[key] = (x, y)
            ws.input_rhs_indices = np.array((tuple(ins)), dtype=int)
            ws.output_rhs_indices = np.array((tuple(outs)), dtype=int)
            ws.out_fsig_sweep = np.zeros(
                (len(ins), len(self.f), len(outs)), dtype=np.complex128
            )
        except AttributeError:
            raise Exception(
                "Noise analysis requires a signal simulation to work, please set the model fsig.f frequency"
            )

        return ws

    def update_symbolic_values(self, ws):
        run_fsig_sweep(
            ws.carrier,
            ws.signal,
            self.f,
            ws.input_rhs_indices,
            ws.output_rhs_indices,
            ws.out_fsig_sweep,
            True,
        )
        for sym in ws.syms:
            if sym is ws.Omega:
                continue
            el = ws.model.elements.get(str(sym))
            if el:
                ws.values[sym] = el.eval(ws.OMEGA)
            else:
                # Now we loop over the actual simulation and run each point
                i, o = ws.optical_couplings[sym]
                ws.values[sym] = ws.out_fsig_sweep[i, :, o]

    def do(self, ws):
        ws.sol = NoiseSolution(
            self.name, self.f, self.output_nodes, ws.noises, ws.s_prev
        )
        ws.sol.Minv = ws.Minv
        ws.sol.Minv_nodes = ws.M_nodes
        self.update_symbolic_values(ws)
        ws.sol.optical_tfs = ws.out_fsig_sweep
        v = ws.values.values()
        for i, onode in enumerate(self.output_nodes):
            ws.sol.noises[onode] = {}
            for j, noise in enumerate(ws.noises):
                y_noise = ws.lambdas[j](self.f) * ws.TF[i][j](*v)
                ws.sol.noises[onode][noise.name] = y_noise



"""Miscellaneous utility functions for any part of Finesse."""

import copy
import logging
import functools
import re
from itertools import tee
from contextlib import contextmanager
from functools import partial

import numpy as np


def calltracker(func):
    """Decorator used for keeping track of whether the current state is inside the
    decorated function or not.

    Sets an attribute `has_been_called` on the function which gets switched on when the
    function is being executed and switched off after the function has returned. This
    allows you to query ``func.has_been_called`` for determining whether the code being
    executed has been called from within `func`.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        wrapper.has_been_called = True
        out = func(*args, **kwargs)
        wrapper.has_been_called = False

        return out

    wrapper.has_been_called = False
    return wrapper


def valid_name(name):
    """Validate the specified name."""
    return re.compile("^[a-zA-Z_][a-zA-Z0-9_]*$").match(name)


def check_name(name):
    """Checks the validity of a component or node name.

    A name is valid if it contains only alphanumeric characters and underscores, and is
    not empty.

    Parameters
    ----------
    name : str
        The name to check.

    Returns
    -------
    name : str
        The name passed to this function if valid.

    Raises
    ------
    ValueError
        If `name` contains non-alphanumeric / underscore characters.
    """
    if not valid_name(name):
        raise ValueError(
            "Name `{}` is not valid. Alphanumeric and underscores only".format(name)
        )
    return name


def pairwise(iterable):
    """Iterates through each pair in a iterable.

    Parameters
    ----------
    iterable : :py:class:`collections.abc.Iterable`
        An iterable object.

    Returns
    -------
    zip
        A zip object whose `.next()` method returns a tuple where the i-th
        element comes from the i-th iterable argument.
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


def find(x, value):
    """Finds `value` in the list `x` and returns its index, returning `None` if `value`
    is not in the list."""
    try:
        return x.index(value)
    except ValueError:
        return None


def find_nearest(x, value, index=False):
    idx = np.argmin(np.abs(x - value))
    if index:
        return idx
    return x[idx]


def is_iterable(obj):
    """Reliable check for whether an object is iterable.

    Note that strings are treated as non-iterable objects
    when performing this check. This will only return true
    for iterable non-str objects.

    Returns
    -------
    flag : bool
        True if `obj` is iterable, False otherwise.
    """
    try:
        iter(obj)
    except Exception:
        return False
    else:
        return not isinstance(obj, str)


@contextmanager
def changed_params(subs: dict):
    r"""Temporarily change model parameters within a context.

    This is a convenience function for providing a pattern of
    changing parameters temporarily, inside a context, without
    needing to make deep-copies of a :class:`.Model`. Usage of
    this function should be confined to ``with`` contexts.

    Parameters
    ----------
    subs : dict
        Dictionary of temporary parameter subsitutions. Expects a dict
        of :class:`.Parameter` to float value mappings.

    Examples
    --------
    In this example a simple Fabry-Perot cavity is created and
    this function is used to inspect how some cavity properties
    vary when temporarily changing parameters.

    .. jupyter-execute::

        import finesse
        import finesse.components as components

        from finesse.utilities import changed_params

        IFO = finesse.Model()
        IFO.chain(
            components.Laser("L0"),
            components.Mirror("ITM", Rc=-10),
            components.Space("CAV", L=1),
            components.Mirror("ETM", Rc=10),
        )
        IFO.add(components.Cavity("FP", IFO.ITM.p2))

        print(f"Initial FP.g = {IFO.FP.g}")
        print(f"Initial FP.gouy = {IFO.FP.gouy}\n")

        # Temporarily change ITM.Rcx to -15 m and ETM.Rcy to 9 m
        with changed_params({IFO.ITM.Rcx: -15, IFO.ETM.Rcy: 9}):
            print(f"Temporary FP.g = {IFO.FP.g}")
            print(f"Temporary FP.gouy = {IFO.FP.gouy}\n")

        # Check that params have been reset by looking at cavity again
        print(f"Final FP.g = {IFO.FP.g}")
        print(f"Final FP.gouy = {IFO.FP.gouy}")
    """
    # Get the current values of the parameters being changed
    initial_values = {p: copy.copy(p.value) for p in subs}
    # Set the parameters to their temp values
    for p, tval in subs.items():
        p.value = tval

    yield

    # Reset the values on exit
    for p in subs:
        p.value = initial_values[p]


@contextmanager
def opened_file(fobj, mode):
    """Get an open file regardless of whether a string or an already open file is
    passed.

    Parameters
    ----------
    fobj : str or file-like
        The path or file object to ensure is open. If `fobj` is an already open file object, its
        mode is checked to be correct but is otherwise returned as-is. If `fobj` is a string, it is
        opened with the specified `mode` and yielded, then closed once the wrapped context exits.
        Note that passed open file objects are *not* closed.

    mode : str
        The mode to ensure `fobj` is opened with.

    Yields
    ------
    :py:class:`io.FileIO`
        The open file with the specified `mode`.

    Raises
    ------
    ValueError
        If `fobj` is not a string nor open file, or if `fobj` is open but with a different `mode`.
    """
    close = False

    if isinstance(fobj, str):
        fobj = open(fobj, mode)
        close = True  # Close the file we just opened once we're done.
    else:
        try:
            # Ensure mode agrees.
            if fobj.mode != mode:
                raise ValueError(
                    f"Unexpected mode for '{fobj.name}' (expected '{mode}', got '{fobj.mode}')"
                )
        except AttributeError:
            raise ValueError(f"'{fobj}' is not an open file or path.")

    try:
        yield fobj
    finally:
        if close:
            fobj.close()


@contextmanager
def logs(level=None, exclude=None, fmt=None, datefmt=None, jupyter_tracebacks=None):
    """Change the default Finesse log stream handler behaviour in the encapsulated
    context.

    Parameters
    ----------
    level : str or int, optional
        The minimum log levels to print. The standard log levels "debug", "info",
        "warning", "error" and "critical" are supported, as are their corresponding
        level numbers (see :py:mod:`logging`).

    exclude : str or list of str, optional
        Log channel(s) to exclude. Simple wildcards are supported:

        - ``*`` matches 0 or more characters
        - ``?`` matches any single character
        - ``[abc]`` matches any characters in abc
        - ``[!abc]`` matches any characters not in abc

    fmt, datefmt : str, optional
        The stream format to use. The string specified for `fmt` should use "{" style
        parameters as supported by :class:`logging.Formatter`. The `datefmt` string
        should be a :func:`time.strftime` compatible format. Defaults to the existing
        builtin stream handler's default format.

    jupyter_tracebacks : bool, optional
        Whether to show full Jupyter tracebacks for :class:`parsing errors
        <.KatParserError>` raised within the encapsulated context. Defaults to `None`,
        which leaves the current value (which is by default determined by the
        environment in which Finesse is running) unchanged.

    Raises
    ------
    :class:`.ExistingLogHandlerError`
        If the current :mod:`finesse` root logger's first handler was not the one
        autoconfigured by Finesse upon first import. This function does not handle cases
        where the user has configured their own logging.

    Examples
    --------
    Print debug logs during parsing.

    >>> from finesse import Model
    >>> from finesse.utilities import logs
    >>> model = Model()
    >>> with logs(level="debug"):
    >>>     model.parse("laser l1 P=1")
    """
    from .. import show_tracebacks, PROGRAM
    from ..config import log_handler_instance

    logger = logging.getLogger(PROGRAM)
    handler = log_handler_instance()
    old_level = None
    old_excludes = None
    old_formatter = None
    old_tracebacks = None

    if level is not None:
        old_level = logger.level

        try:
            # Convert level name to uppercase.
            level = level.upper()
        except AttributeError:
            # Probably a number.
            pass

        logger.setLevel(level)

    if exclude is not None:
        if isinstance(exclude, str):
            exclude = [exclude]

        old_excludes = handler.reset_exclude_patterns()
        for pattern in exclude:
            handler.exclude(pattern)

    if fmt is not None or datefmt is not None:
        # We can't copy the format string out of the formatter, so we replace the object
        # instead and restore it later.
        old_formatter = handler.formatter

        if datefmt is None:
            # Use the default date format from the old handler.
            datefmt = old_formatter.datefmt

        handler.setFormatter(logging.Formatter(fmt, datefmt, style="{"))

    if jupyter_tracebacks is not None:
        old_tracebacks = show_tracebacks(jupyter_tracebacks)

    # Run the code in the context.
    yield

    if old_level is not None:
        logger.setLevel(old_level)

    if old_excludes is not None:
        handler.reset_exclude_patterns()
        for pattern in old_excludes:
            handler.exclude(pattern)

    if old_formatter is not None:
        handler.setFormatter(old_formatter)

    if old_tracebacks is not None:
        show_tracebacks(old_tracebacks)


def graph_layouts():
    """Available NetworkX and graphviz (if installed) graph plotting layouts."""
    return {**networkx_layouts(), **graphviz_layouts()}


def networkx_layouts():
    """Available NetworkX graph plotting layouts."""
    import inspect
    import networkx

    # Excluded layouts.
    excluded = (
        "rescale",
        # These layouts need the network to first be grouped into sets.
        "bipartite",
        "multipartite_layout",
    )

    layouts = {}
    suffix = "_layout"

    def find_layouts(members):
        for name, func in members:
            if name.startswith("_") or not name.endswith(suffix):
                # Not a public layout.
                continue

            # Strip the "_layout" part.
            name = name[: -len(suffix)]

            if name in excluded:
                continue

            layouts[name] = func

    find_layouts(inspect.getmembers(networkx.drawing.layout, inspect.isfunction))

    return layouts


def graphviz_layouts():
    """Available graphviz graph plotting layouts."""
    import networkx
    from ..environment import has_pygraphviz

    layouts = {}

    if has_pygraphviz():
        for layout in ["neato", "dot", "fdp", "sfdp", "circo"]:
            # Returns callable that can be called like `networkx.drawing.layout` members.
            layouts[layout] = partial(
                networkx.drawing.nx_agraph.pygraphviz_layout, prog=layout
            )

    return layouts


def doc_element_parameter_table(cls):
    """Prints table for a particular element class."""
    import finesse
    import tabulate
    from IPython.core.display import HTML
    from functools import partial
    from types import SimpleNamespace

    # object counter to keep track of odd and even rows
    i = SimpleNamespace()
    i.current = -1

    def my_html_row_with_attrs(celltag, cell_values, colwidths, colaligns, **kwargs):
        alignment = {
            "left": ' style="text-align: left;"',
            "right": ' style="text-align: right;"',
            "center": ' style="text-align: center;"',
            "decimal": ' style="text-align: right;"',
        }
        values_with_attrs = [
            "<{0}{1}>{2}</{0}>".format(celltag, alignment.get(a, ""), c)
            for c, a in zip(cell_values, colaligns)
        ]
        kwargs["i"].current += 1
        return (
            f"<tr style=\"vertical-align: middle;\" class=\"row-{'even' if kwargs['i'].current % 2 else 'odd'}\">"
            + "".join(values_with_attrs).rstrip()
            + "</tr>"
        )

    def process_changeable(pinfo):
        if pinfo.changeable_during_simulation:
            return "<div style='color:green;'>✓</div>"
        else:
            return "<div style='color:red;'>✗</div>"

    SphinxTable = tabulate.TableFormat(
        lineabove=tabulate.Line(
            '<table class="docutils align-default"><thead>', "", "", ""
        ),
        linebelowheader=tabulate.Line("</thead>", "", "", ""),
        linebetweenrows=None,
        linebelow=tabulate.Line("</table>", "", "", ""),
        headerrow=partial(my_html_row_with_attrs, "th", i=i),
        datarow=partial(my_html_row_with_attrs, "td", i=i),
        padding=0,
        with_header_hide=None,
    )
    tbl = [
        (p.name, p.description, p.units, p.dtype.__name__, process_changeable(p))
        for p in finesse.element.ModelElement._param_dict[cls][::-1]
    ]
    a = tabulate.tabulate(
        tbl,
        tablefmt=SphinxTable,
        colalign=("left", "left", "center", "center", "center"),
        headers=(
            "Name",
            "Description",
            "Units",
            "Data type",
            "Can change during simualation",
        ),
    )
    return HTML(a)



"""Tests for the tokenizer without any normalization."""

import re
import pytest
from finesse.script.exceptions import KatScriptError


@pytest.mark.parametrize(
    "string,error",
    (
        ("[", ("\nline 1: unclosed '['\n" "-->1: [\n" "      ^")),
        ("(", ("\nline 1: unclosed '('\n" "-->1: (\n" "      ^")),
        ("]", ("\nline 1: extraneous ']'\n" "-->1: ]\n" "      ^")),
        (")", ("\nline 1: extraneous ')'\n" "-->1: )\n" "      ^")),
    ),
)
def test_invalid_delimiter_nesting(tokenizer, string, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        list(tokenizer.tokenize(string))



"""Parsing and unparsing of Finesse kat files and models."""

from ..utilities import opened_file


def parse(text, model=None, spec=None):
    """Parse kat script into a model.

    Parameters
    ----------
    text : str
        The kat script to parse.

    model : :class:`.Model`, optional
        The Finesse model to add the parsed objects to. Defaults to a new, empty model.

    spec : :class:`.spec.BaseSpec`, optional
        The language specification to use. Defaults to :class:`.spec.KatSpec`.

    Returns
    -------
    :class:`.Model`
        The parsed model.
    """
    from .compiler import KatCompiler

    compiler = KatCompiler(spec=spec)
    return compiler.compile(text, model=model)


def parse_file(path, model=None, spec=None):
    """Parse kat script from a file into a model.

    Parameters
    ----------
    path : str or :py:class:`io.FileIO`
        The path or file object to read kat script from. If an open file object is
        passed, it will be read from and left open. If a path is passed, it will be
        opened, read from, then closed.

    model : :class:`.Model`, optional
        The Finesse model to add the parsed objects to. Defaults to a new, empty model.

    spec : :class:`.spec.BaseSpec`, optional
        The language specification to use. Defaults to :class:`.spec.KatSpec`.

    Returns
    -------
    :class:`.Model`
        The parsed model.
    """
    from .compiler import KatCompiler

    compiler = KatCompiler(spec=spec)
    with opened_file(path, "r") as fobj:
        return compiler.compile_file(fobj, model=model)


def parse_legacy(text, model=None, ignored_blocks=None):
    """Parse kat script into a model.

    Parameters
    ----------
    text : str
        The kat script to parse.

    model : :class:`.Model`
        The Finesse model to add the parsed objects to.

    ignored_blocks : list, optional
        A list of names of ``FTBLOCK`` sections in the kat code to leave out of the
        model; defaults to empty list.

    Returns
    -------
    :class:`.Model`
        The parsed model.

    Raises
    ------
    NotImplementedError
        If `model` contains any non-default elements. Parsing into existing models is
        unsupported.
    """
    from .legacy import KatParser

    if model:
        # Newly-created models contain an fsig, so we need to account for that
        if len(model.elements) > 1:
            raise NotImplementedError(
                "Legacy parsing of extra commands with an existing model is "
                "unsupported. Please switch to the new syntax, or only call "
                "'parse_legacy' on a complete kat file."
            )
    parser = KatParser()
    return parser.parse(text, model=model, ignored_blocks=ignored_blocks)


def parse_legacy_file(path, model=None, ignored_blocks=None):
    """Parse kat script from a file into a model.

    Parameters
    ----------
    path : str or :py:class:`io.FileIO`
        The path or file object to read kat script from. If an open file object is
        passed, it will be read from and left open. If a path is passed, it will be
        opened, read from, then closed.

    model : :class:`.Model`
        The Finesse model to add the parsed objects to.

    ignored_blocks : list, optional
        A list of names of ``FTBLOCK`` sections in the kat code to leave out of the
        model; defaults to empty list.

    Returns
    -------
    :class:`.Model`
        The parsed model.

    Raises
    ------
    NotImplementedError
        If `model` contains any non-default elements. Parsing into existing models is
        unsupported.
    """
    from .legacy import KatParser

    if model:
        # Newly-created models contain an fsig, so we need to account for that
        if len(model.elements) > 1:
            raise NotImplementedError(
                "Legacy parsing of extra commands with an existing model is "
                "unsupported. Please switch to the new syntax, or only call "
                "'parse_legacy' on a complete kat file."
            )

    parser = KatParser()
    with opened_file(path, "r") as fobj:
        return parser.parse(fobj.read(), model=model, ignored_blocks=ignored_blocks)


def unparse(model):
    """Serialise a model to kat script.

    Parameters
    ----------
    model : :class:`.Model`
        The Finesse model to generate kat script for.

    Returns
    -------
    str
        The generated kat script.
    """
    from .generator import KatUnbuilder

    unbuilder = KatUnbuilder()
    return unbuilder.unbuild(model)


def unparse_file(path, model):
    """Serialise a model to kat script in a file.

    Parameters
    ----------
    path : str
        The kat file path to parse.

    model : :class:`.Model`
        The Finesse model to generate kat script for.

    Returns
    -------
    str
        The generated kat script.
    """
    from .generator import KatUnbuilder

    unbuilder = KatUnbuilder()
    with opened_file(path, "w") as fobj:
        return unbuilder.unbuild_file(fobj, model)


def syntax(directive, spec=None, **kwargs):
    """Get the syntax for `directive`.

    Parameters
    ----------
    directive : str
        The directive to retrieve syntax for.

    spec : :class:`.KatSpec`, optional
        The kat script specification to use. Defaults to :class:`.KatSpec`.

    Other Parameters
    ----------------
    kwargs : dict, optional
        Keyword arguments supported by :meth:`.KatSyntaxUnparser.syntax`.

    Returns
    -------
    str
        The syntax for `directive`.
    """
    from .generator import KatSyntaxUnparser

    unparser = KatSyntaxUnparser(spec=spec)
    return unparser.syntax(directive, **kwargs)


__all__ = (
    "KatParserError",
    "KatReferenceError",
    "parse",
    "parse_file",
    "parse_legacy",
    "parse_legacy_file",
    "unparse",
    "unparse_file",
    "syntax",
)



import pytest
from finesse.script.containers import KatScript, KatFunction, KatKwarg
from ...util import dedent_multiline
from ..util import (
    SPACE,
    COMMA,
    NEWLINE,
    NAME,
    NUMBER,
    EQUALS,
    LPAREN,
    RPAREN,
    COMMENT,
)


@pytest.mark.parametrize(
    "script,function",
    (
        (
            # Function with 4 positional arguments.
            "fake_function(1, 2, 3, 4)",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[
                    NUMBER(1, 15, "1"),
                    NUMBER(1, 18, "2"),
                    NUMBER(1, 21, "3"),
                    NUMBER(1, 24, "4"),
                ],
                extra=[
                    LPAREN(1, 14),
                    COMMA(1, 16),
                    SPACE(1, 17),
                    COMMA(1, 19),
                    SPACE(1, 20),
                    COMMA(1, 22),
                    SPACE(1, 23),
                    RPAREN(1, 25),
                ],
            ),
        ),
        (
            # Function with 2 positional arguments and 2 keyword arguments.
            "fake_function(1, 2, c=3, d=4)",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[
                    NUMBER(1, 15, "1"),
                    NUMBER(1, 18, "2"),
                    KatKwarg(
                        key=NAME(1, 21, "c"),
                        value=NUMBER(1, 23, "3"),
                        extra=[EQUALS(1, 22)],
                    ),
                    KatKwarg(
                        key=NAME(1, 26, "d"),
                        value=NUMBER(1, 28, "4"),
                        extra=[EQUALS(1, 27)],
                    ),
                ],
                extra=[
                    LPAREN(1, 14),
                    COMMA(1, 16),
                    SPACE(1, 17),
                    COMMA(1, 19),
                    SPACE(1, 20),
                    COMMA(1, 24),
                    SPACE(1, 25),
                    RPAREN(1, 29),
                ],
            ),
        ),
        (
            # Function with 4 keyword arguments.
            "fake_function(a=1, b=2, c=3, d=4)",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[
                    KatKwarg(
                        key=NAME(1, 15, "a"),
                        value=NUMBER(1, 17, "1"),
                        extra=[EQUALS(1, 16)],
                    ),
                    KatKwarg(
                        key=NAME(1, 20, "b"),
                        value=NUMBER(1, 22, "2"),
                        extra=[EQUALS(1, 21)],
                    ),
                    KatKwarg(
                        key=NAME(1, 25, "c"),
                        value=NUMBER(1, 27, "3"),
                        extra=[EQUALS(1, 26)],
                    ),
                    KatKwarg(
                        key=NAME(1, 30, "d"),
                        value=NUMBER(1, 32, "4"),
                        extra=[EQUALS(1, 31)],
                    ),
                ],
                extra=[
                    LPAREN(1, 14),
                    COMMA(1, 18),
                    SPACE(1, 19),
                    COMMA(1, 23),
                    SPACE(1, 24),
                    COMMA(1, 28),
                    SPACE(1, 29),
                    RPAREN(1, 33),
                ],
            ),
        ),
        # Trailing comma.
        pytest.param(
            "fake_function(1,)",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[NUMBER(1, 15, "1"),],
                extra=[LPAREN(1, 14), COMMA(1, 16), RPAREN(1, 17),],
            ),
            id="single-line-trailing-comma-arg-nospace",
        ),
        pytest.param(
            "fake_function(1 ,)",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[NUMBER(1, 15, "1"),],
                extra=[LPAREN(1, 14), SPACE(1, 16), COMMA(1, 17), RPAREN(1, 18),],
            ),
            id="single-line-trailing-comma-arg-prespace",
        ),
        pytest.param(
            "fake_function(1, )",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[NUMBER(1, 15, "1"),],
                extra=[LPAREN(1, 14), COMMA(1, 16), SPACE(1, 17), RPAREN(1, 18),],
            ),
            id="single-line-trailing-comma-arg-postspace",
        ),
        pytest.param(
            "fake_function(1 , )",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[NUMBER(1, 15, "1"),],
                extra=[
                    LPAREN(1, 14),
                    SPACE(1, 16),
                    COMMA(1, 17),
                    SPACE(1, 18),
                    RPAREN(1, 19),
                ],
            ),
            id="single-line-trailing-comma-arg-prepostspace",
        ),
        pytest.param(
            "fake_function(a=1,)",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[
                    KatKwarg(
                        key=NAME(1, 15, "a"),
                        value=NUMBER(1, 17, "1"),
                        extra=[EQUALS(1, 16)],
                    ),
                ],
                extra=[LPAREN(1, 14), COMMA(1, 18), RPAREN(1, 19),],
            ),
            id="single-line-trailing-comma-kwarg-nospace",
        ),
        pytest.param(
            "fake_function(a=1 ,)",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[
                    KatKwarg(
                        key=NAME(1, 15, "a"),
                        value=NUMBER(1, 17, "1"),
                        extra=[EQUALS(1, 16)],
                    ),
                ],
                extra=[LPAREN(1, 14), SPACE(1, 18), COMMA(1, 19), RPAREN(1, 20),],
            ),
            id="single-line-trailing-comma-kwarg-prespace",
        ),
        pytest.param(
            "fake_function(a=1, )",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[
                    KatKwarg(
                        key=NAME(1, 15, "a"),
                        value=NUMBER(1, 17, "1"),
                        extra=[EQUALS(1, 16)],
                    ),
                ],
                extra=[LPAREN(1, 14), COMMA(1, 18), SPACE(1, 19), RPAREN(1, 20),],
            ),
            id="single-line-trailing-comma-kwarg-postspace",
        ),
        pytest.param(
            "fake_function(a=1 , )",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[
                    KatKwarg(
                        key=NAME(1, 15, "a"),
                        value=NUMBER(1, 17, "1"),
                        extra=[EQUALS(1, 16)],
                    ),
                ],
                extra=[
                    LPAREN(1, 14),
                    SPACE(1, 18),
                    COMMA(1, 19),
                    SPACE(1, 20),
                    RPAREN(1, 21),
                ],
            ),
            id="single-line-trailing-comma-kwarg-prepostspace",
        ),
        pytest.param(
            "fake_function(1, b=2 , )",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[
                    NUMBER(1, 15, "1"),
                    KatKwarg(
                        key=NAME(1, 18, "b"),
                        value=NUMBER(1, 20, "2"),
                        extra=[EQUALS(1, 19)],
                    ),
                ],
                extra=[
                    LPAREN(1, 14),
                    COMMA(1, 16),
                    SPACE(1, 17),
                    SPACE(1, 21),
                    COMMA(1, 22),
                    SPACE(1, 23),
                    RPAREN(1, 24),
                ],
            ),
            id="single-line-trailing-comma-arg-kwarg-prepostspace",
        ),
    ),
)
def test_single_line(parser, script, function):
    """Single line function."""
    assert parser.parse(script) == KatScript(arguments=[function], extra=[])


@pytest.mark.parametrize(
    "script,function",
    (
        (
            dedent_multiline(
                """
                fake_function(
                    1, 2, 3, 4
                )
                """
            ),
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[
                    NUMBER(2, 5, "1"),
                    NUMBER(2, 8, "2"),
                    NUMBER(2, 11, "3"),
                    NUMBER(2, 14, "4"),
                ],
                extra=[
                    LPAREN(1, 14),
                    NEWLINE(1, 15),
                    SPACE(2, 1, 4),
                    COMMA(2, 6),
                    SPACE(2, 7),
                    COMMA(2, 9),
                    SPACE(2, 10),
                    COMMA(2, 12),
                    SPACE(2, 13),
                    NEWLINE(2, 15),
                    RPAREN(3, 1),
                ],
            ),
        ),
        (
            dedent_multiline(
                """
                fake_function(
                    1,
                    2,
                    a=3,
                    b=4
                )
                """
            ),
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[
                    NUMBER(2, 5, "1"),
                    NUMBER(3, 5, "2"),
                    KatKwarg(
                        key=NAME(4, 5, "a"),
                        value=NUMBER(4, 7, "3"),
                        extra=[EQUALS(4, 6)],
                    ),
                    KatKwarg(
                        key=NAME(5, 5, "b"),
                        value=NUMBER(5, 7, "4"),
                        extra=[EQUALS(5, 6)],
                    ),
                ],
                extra=[
                    LPAREN(1, 14),
                    NEWLINE(1, 15),
                    SPACE(2, 1, 4),
                    COMMA(2, 6),
                    NEWLINE(2, 7),
                    SPACE(3, 1, 4),
                    COMMA(3, 6),
                    NEWLINE(3, 7),
                    SPACE(4, 1, 4),
                    COMMA(4, 8),
                    NEWLINE(4, 9),
                    SPACE(5, 1, 4),
                    NEWLINE(5, 8),
                    RPAREN(6, 1),
                ],
            ),
        ),
        # Trailing comma.
        pytest.param(
            dedent_multiline(
                """
                fake_function(
                    1,
                    # 2,
                )
                """
            ),
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[NUMBER(2, 5, "1")],
                extra=[
                    LPAREN(1, 14),
                    NEWLINE(1, 15),
                    SPACE(2, 1, 4),
                    COMMA(2, 6),
                    NEWLINE(2, 7),
                    SPACE(3, 1, 4),
                    COMMENT(3, 5, "# 2,"),
                    NEWLINE(3, 9),
                    RPAREN(4, 1),
                ],
            ),
            id="multi-line-trailing-comma",
        ),
    ),
)
def test_multi_line(parser, script, function):
    """Multi-line function."""
    assert parser.parse(script) == KatScript(arguments=[function], extra=[])


@pytest.mark.parametrize(
    "script,function",
    (
        (
            "fake_function()",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[],
                extra=[LPAREN(1, 14), RPAREN(1, 15)],
            ),
        ),
        (
            "fake_function( )",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[],
                extra=[LPAREN(1, 14), SPACE(1, 15), RPAREN(1, 16)],
            ),
        ),
        (
            "fake_function(  )",
            KatFunction(
                directive=NAME(1, 1, "fake_function"),
                arguments=[],
                extra=[LPAREN(1, 14), SPACE(1, 15, 2), RPAREN(1, 17)],
            ),
        ),
    ),
)
def test_empty_arguments(parser, script, function):
    assert parser.parse(script) == KatScript(arguments=[function], extra=[])


@pytest.mark.parametrize(
    "script,function",
    (
        (
            # Nested statement without parameters.
            dedent_multiline(
                """
                fake_analysis(
                    nested1(
                        1, 2
                    ),
                    nested2(
                        3, 4
                    ),
                    on_complete=subanalysis()
                )
                """
            ),
            KatFunction(
                directive=NAME(1, 1, "fake_analysis"),
                arguments=[
                    KatFunction(
                        directive=NAME(2, 5, "nested1"),
                        arguments=[NUMBER(3, 9, "1"), NUMBER(3, 12, "2"),],
                        extra=[
                            LPAREN(2, 12),
                            NEWLINE(2, 13),
                            SPACE(3, 1, 8),
                            COMMA(3, 10),
                            SPACE(3, 11),
                            NEWLINE(3, 13),
                            SPACE(4, 1, 4),
                            RPAREN(4, 5),
                        ],
                    ),
                    KatFunction(
                        directive=NAME(5, 5, "nested2"),
                        arguments=[NUMBER(6, 9, "3"), NUMBER(6, 12, "4"),],
                        extra=[
                            LPAREN(5, 12),
                            NEWLINE(5, 13),
                            SPACE(6, 1, 8),
                            COMMA(6, 10),
                            SPACE(6, 11),
                            NEWLINE(6, 13),
                            SPACE(7, 1, 4),
                            RPAREN(7, 5),
                        ],
                    ),
                    KatKwarg(
                        key=NAME(8, 5, "on_complete"),
                        value=KatFunction(
                            directive=NAME(8, 17, "subanalysis"),
                            arguments=[],
                            extra=[LPAREN(8, 28), RPAREN(8, 29)],
                        ),
                        extra=[EQUALS(8, 16)],
                    ),
                ],
                extra=[
                    LPAREN(1, 14),
                    NEWLINE(1, 15),
                    SPACE(2, 1, 4),
                    COMMA(4, 6),
                    NEWLINE(4, 7),
                    SPACE(5, 1, 4),
                    COMMA(7, 6),
                    NEWLINE(7, 7),
                    SPACE(8, 1, 4),
                    NEWLINE(8, 30),
                    RPAREN(9, 1),
                ],
            ),
        ),
    ),
)
def test_nested(parser, script, function):
    assert parser.parse(script) == KatScript(arguments=[function], extra=[])



"""Model component network unit test fixtures."""

import pytest
from finesse import Model
from finesse.components import Laser, Beamsplitter, Mirror, Cavity
from finesse.detectors import PowerDetector


@pytest.fixture(scope="module")
def network_model_michelson():
    """Fixture with a Michelson model to test component network generation."""
    model = Model()

    model.add(
        [
            Laser("L0"),
            Beamsplitter("BS"),
            Mirror("IMX"),
            Mirror("IMY"),
            Mirror("EMX"),
            Mirror("EMY"),
        ]
    )
    model.connect(model.L0.p1, model.BS.p1)
    model.connect(model.BS.p2, model.IMY.p1)
    model.connect(model.IMY.p2, model.EMY.p1)
    model.connect(model.BS.p3, model.IMX.p1)
    model.connect(model.IMX.p2, model.EMX.p1)

    model.add(PowerDetector("refl", model.IMX.p1.o))
    model.add(Cavity("cav_x", model.IMX.p2.o, model.IMX.p2.i))
    model.add(Cavity("cav_y", model.IMY.p2.o, model.IMY.p2.i))

    return model, model.to_component_network()


@pytest.fixture(scope="module")
def network_model_sagnac():
    """Fixture with a Sagnac model to test component network generation."""
    model = Model()

    model.add(
        [Laser("L0"), Beamsplitter("BS"), Mirror("M1"), Mirror("M2"), Mirror("M3")]
    )
    model.connect(model.L0.p1, model.BS.p1)
    model.connect(model.BS.p2, model.M1.p1)
    model.connect(model.M1.p2, model.M2.p1)
    model.connect(model.M2.p2, model.M3.p1)
    model.connect(model.M3.p2, model.BS.p3)

    model.add(PowerDetector("trns", model.BS.p4.o))

    return model, model.to_component_network()



"""Some fast C level functions implemented in finesse.cymath use explicitly hard-coded
values or look-up tables.

These are tested against general implementations (from scipy) here.
"""

import numpy as np
from numpy.testing import assert_allclose
import pytest

import scipy.special as scs
import finesse.cymath as cymath


@pytest.mark.parametrize("x", (np.pi, -np.pi, 1e-9, 1e12))
def test_explicit_fast_hermite_function(x):
    """Test that hard-coded (up to n=20) fast C hermite implementation cases are
    correct.

    For n > 20, the recursive relation is used in the cymath implementation so no need
    to test above this limit.
    """
    ns = np.arange(21)

    expected = scs.eval_hermite(ns, x)
    computed = [cymath.math.hermite(n, x) for n in ns]

    assert_allclose(expected, computed)


def test_factorial_lookup_table():
    """Test that look-up table for factorials is correct up to n = 20."""
    ns = np.arange(21)

    expected = scs.factorial(ns)
    computed = [cymath.math.factorial(n) for n in ns]

    assert_allclose(expected, computed)


def test_sqrt_factorial_lookup_table():
    """Test that look-up table for sqrt(factorials) is correct up to n = 20."""
    ns = np.arange(21)

    expected = np.sqrt(scs.factorial(ns))
    computed = [cymath.math.sqrt_factorial(n) for n in ns]

    assert_allclose(expected, computed)



"""Kat file parser reference syntax tests."""

import re
import pytest
from finesse.script import parse
from finesse.script.exceptions import KatScriptError
from ...util import dedent_multiline


@pytest.fixture
def model_with_references(model):
    model.parse(
        """
        var L 1
        pd P ITM.p1.o

        l L0 P=1
        s s1 L0.p1 ITM.p1
        m ITM R=1-&ETM.T L=0
        s sCAV ITM.p2 ETM.p1 L=&L
        m ETM R=0.99 T=0.01
        """
    )

    return model


@pytest.fixture
def model_with_references_second_parse(model_with_references):
    model_with_references.parse("xaxis(L0.P, lin, 0, 100, 10)")
    return model_with_references


def test_reference():
    """Test parameter reference syntax."""
    parse(
        """
        l l1
        l l2 P=&l1.P
        """
    )


def test_default_reference():
    """Test that references to elements (and not their model parameters) return their default."""
    model = parse(
        """
        var myvar 180
        l l1
        l l2 P=&l1.P phase=&myvar
        """
    )
    assert model.l2.phase.eval() == 180


@pytest.mark.parametrize(
    "script,error",
    (
        pytest.param(
            dedent_multiline(
                """
                l l1
                l l2 P=&l1.P phase=&l1
                """
            ),
            (
                "\nline 2: l1 cannot be referenced because type 'Laser' has no default model "
                "parameter (hint: try '&l1.[some parameter]')\n"
                "   1: l l1\n"
                "-->2: l l2 P=&l1.P phase=&l1\n"
                "                         ^^^"
            ),
            id="ref-to-element-without-default-parameter",
        ),
        pytest.param(
            "l l1 P=&l2.P",
            (
                "\nline 1: model has no attribute 'l2.P'\n"
                "-->1: l l1 P=&l2.P\n"
                "             ^^^^^"
            ),
            id="ref-to-non-existent-element",
        ),
        pytest.param(
            "l l1 P=&l2.P.",
            (
                "\nline 1: 'l2.P.' should not end with a '.'\n"
                "-->1: l l1 P=&l2.P.\n"
                "             ^^^^^^"
            ),
            id="ref-to-non-existent-element",
        ),
        pytest.param(
            "l l1 P=&l1.P",
            (
                "\nline 1: cannot set l1.P to self-referencing value l1.P\n"
                "-->1: l l1 P=&l1.P\n"
                "             ^^^^^"
            ),
            id="ref-to-self",
        ),
    ),
)
def test_invalid_reference(script, error):
    with pytest.raises(KatScriptError, match=re.escape(error)):
        parse(script)


def test_references(model_with_references):
    assert model_with_references.ITM.R.eval() == 1 - model_with_references.ETM.T.eval()
    assert model_with_references.spaces.sCAV.L.eval() == 1


def test_first_run(model_with_references):
    model_with_references.run()


def test_second_parse_run(model_with_references_second_parse):
    model_with_references_second_parse.run()


def test_geometric_parameters_as_references():
    """Test a model containing a geometric model parameter referencing another element's value.

    See #172.
    """
    model = parse(
        """
        l L0 P=1

        s s0 L0.p1 ITM.p1

        m ITM R=0.99 T=0.01 Rc=-10
        s CAV ITM.p2 ETM.p1 L=&L    # <-------
        m ETM R=0.99 T=0.01 Rc=10

        cav FP ITM.p2.o

        var L 1
        """
    )

    # Check reference.
    assert model.spaces.CAV.L.value == model.L.value
    # Check value.
    assert model.spaces.CAV.L.eval() == 1
